"""Navigation elements.

Abilian define theses categories:   `section`:     Used for navigation
elements relevant to site section   `user`:     User for element that
should appear in user menu
"""
import typing
from typing import Any, Dict, Optional, Tuple, Union

from flask import g
from flask_babel.speaklater import LazyString
from jinja2 import Markup, Template

from .action import ACTIVE, ENABLED, Action, Glyphicon, getset

if typing.TYPE_CHECKING:
    from abilian.web.action import Status


class NavItem(Action):
    """A single navigation item."""

    divider = False

    def __init__(
        self, category: str, name: str, divider: bool = False, *args: Any, **kwargs: Any
    ) -> None:
        category = "navigation:" + category
        super().__init__(category, name, *args, **kwargs)
        self.divider = divider

    @getset
    def status(self, value: Optional[Any] = None) -> "Status":
        current = g.nav.get("active")
        if current is None:
            return ENABLED

        if not current.startswith("navigation:"):
            current = "navigation:" + current

        status = ACTIVE if current == self.path else ENABLED
        return status

    @property
    def path(self) -> str:
        return self.category + ":" + self.name


class NavGroup(NavItem):
    """A navigation group renders a list of items."""

    template_string = """
    <ul class="nav navbar-nav {{ action.css_class }}">
      <li class="dropdown">
        <a class="dropdown-toggle" data-toggle="dropdown">
          {%- if action.icon %}{{ action.icon }}{% endif %}
          {{ action.title }} <b class="caret"></b>
        </a>
        <ul class="dropdown-menu">
          {%- for item in action_items %}
          {%- if item.divider %}<li class="divider"></li>{%- endif %}
          <li class="{{ item.status|safe }}">{{ item.render() }}</li>
          {%- endfor %}
        </ul>
      </li>
    </ul>
    """

    def __init__(
        self, category: str, name: str, items: Tuple[()] = (), *args: Any, **kwargs: Any
    ) -> None:
        NavItem.__init__(self, category, name, *args, **kwargs)
        self.items = list(items)
        self._paths = {self.path}
        for i in self.items:
            self._paths.add(i.path)

    def append(self, item: NavItem) -> None:
        self.items.append(item)
        self._paths.add(item.path)

    def insert(self, pos, item):
        self.items.insert(pos, item)
        self._paths.add(item.path)

    def get_render_args(self, **kwargs: Any) -> Dict[str, Any]:
        params = super().get_render_args(**kwargs)
        params["action_items"] = [a for a in self.items if a.available(params)]
        return params

    @getset
    def status(self, value: Optional[Any] = None) -> "Status":
        current = g.nav.get("active")
        if current is None:
            return ENABLED

        if not current.startswith("navigation:"):
            current = "navigation:" + current
        status = ACTIVE if current in self._paths else ENABLED
        return status


class BreadcrumbItem:
    """A breadcrumb element has at least a label or an icon."""

    #: Label shown to user. May be an i18n string instance
    label = None

    #: Icon to use.
    icon = None

    #: Additional text, can be used as tooltip for example
    description = None

    #: either an Unicode string or an :class:`Endpoint` instance.
    _url = None

    template_string = (
        '{%- if url %}<a href="{{ url }}">{%- endif %}'
        "{%- if item.icon %}{{ item.icon }} {%- endif %}"
        "{{ item.label }}"
        "{%- if url %}</a>{%- endif %}"
    )

    def __init__(
        self,
        label: Union[LazyString, str] = "",
        url="#",
        icon: Optional[str] = None,
        description: Optional[Any] = None,
    ) -> None:
        # don't test 'label or...': if label is a lazy_gettext, it will be
        # resolved. If this item is created in a url_value_preprocessor, it will
        # setup i18n before auth has loaded user, so i18n will fallback on browser
        # negociation instead of user's site preference, and load wrong catalogs for
        # the whole request.
        assert label is not None or icon is None
        self.label = label
        if isinstance(icon, str):
            icon = Glyphicon(icon)

        self.icon = icon
        self.description = description
        self._url = url
        self.__template = Template(self.template_string)

    @property
    def url(self) -> str:
        return str(self._url)

    def render(self):
        return Markup(self.__template.render(item=self, url=self.url))



""""""
import logging
from pprint import pformat

from flask import current_app
from flask.cli import AppGroup

logging.basicConfig()
logger = logging.getLogger("")

config_commands = AppGroup("config")


@config_commands.command()
def show(only_path=False):
    """Show the current config."""
    logger.setLevel(logging.INFO)
    infos = ["\n", f'Instance path: "{current_app.instance_path}"']

    logger.info("\n  ".join(infos))

    if not only_path:
        log_config(current_app.config)


def log_config(config):
    original_level = logger.level
    logger.setLevel(logging.INFO)
    try:
        return _log_config(config)
    finally:
        logger.setLevel(original_level)


def _log_config(config):
    lines = ["Application configuration:"]

    for k, v in sorted(config.items()):
        prefix = "    "
        indent = len(k) + 3
        width = 80 - indent
        v = pformat(v, width=width).replace("\n", "\n" + " " * indent)
        lines.append(f"{prefix}{k}: {v}")
    logger.info("\n".join(lines))



from typing import Any, Dict

from babel.dates import LOCALTZ
from flask import Flask
from werkzeug.datastructures import ImmutableDict

from abilian.web.action import Endpoint


class DefaultConfig:
    # Seriously: this need to be changed in production
    SECRET_KEY = "CHANGEME"

    # Need to be explicitly defined in production configs
    PRODUCTION = False

    # Security (see
    # https://blog.miguelgrinberg.com/post/cookie-security-for-flask-applications)
    # NB: SESSION_COOKIE_* are now set up by Talisman
    WTF_CSRF_ENABLED = True

    # Babel
    BABEL_ACCEPT_LANGUAGES = ["en"]
    DEFAULT_COUNTRY = None

    # Celery
    CELERYD_MAX_TASKS_PER_CHILD = 1000
    CELERY_ACCEPT_CONTENT = ["pickle", "json", "msgpack", "yaml"]
    CELERY_TIMEZONE = LOCALTZ

    # Sentry
    SENTRY_SDK_URL = "https://browser.sentry-cdn.com/4.5.3/bundle.min.js"

    # SQLAlchemy
    SQLALCHEMY_POOL_RECYCLE = 1800  # 30min. default value in flask_sa is None
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    # Debug settings (override default)
    DEBUG_TB_INTERCEPT_REDIRECTS = False

    # Abilian-specific
    PRIVATE_SITE = False
    PLUGINS = ()
    ADMIN_PANELS = (
        "abilian.web.admin.panels.dashboard.DashboardPanel",
        "abilian.web.admin.panels.audit.AuditPanel",
        "abilian.web.admin.panels.login_sessions.LoginSessionsPanel",
        "abilian.web.admin.panels.settings.SettingsPanel",
        "abilian.web.admin.panels.users.UsersPanel",
        "abilian.web.admin.panels.groups.GroupsPanel",
        "abilian.web.admin.panels.sysinfo.SysinfoPanel",
        "abilian.web.admin.panels.impersonate.ImpersonatePanel",
        "abilian.services.vocabularies.admin.VocabularyPanel",
        "abilian.web.tags.admin.TagPanel",
    )
    LOGO_URL = Endpoint("abilian_static", filename="img/logo-abilian-32x32.png")
    ABILIAN_UPSTREAM_INFO_ENABLED = False  # upstream info extension
    TRACKING_CODE = ""  # tracking code for web analytics to insert before </body>
    MAIL_ADDRESS_TAG_CHAR = None


default_config = dict(Flask.default_config)  # type: Dict[str, Any]
default_config.update(vars(DefaultConfig))
default_config = ImmutableDict(default_config)



"""Subject classes (i.e. people, groups, etc.).

See ICOM-ics-v1.0 "Subject Branch".

TODO: I'm not a big fan of the "subject" name. Could be replaced by something
else, like "people" or "principal" ?
"""
import random
import string
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta

import bcrypt
import sqlalchemy as sa
from flask_login import UserMixin
from flask_sqlalchemy import BaseQuery
from sqlalchemy.event import listens_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, deferred, relationship
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.schema import Column, ForeignKey, Table, UniqueConstraint
from sqlalchemy.types import Boolean, DateTime, Integer, LargeBinary, \
    UnicodeText

from abilian.core import sqlalchemy as sa_types
from abilian.core.util import fqcn

from .base import SEARCHABLE, SYSTEM, IdMixin, Indexable, TimestampedMixin, db

__all__ = ["User", "Group", "Principal", "ClearPasswordStrategy", "gen_random_password"]

# Tables for many-to-many relationships
following = Table(
    "following",
    db.Model.metadata,
    Column("follower_id", Integer, ForeignKey("user.id")),
    Column("followee_id", Integer, ForeignKey("user.id")),
    UniqueConstraint("follower_id", "followee_id"),
)

membership = Table(
    "membership",
    db.Model.metadata,
    Column(
        "user_id",
        Integer,
        ForeignKey("user.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    Column(
        "group_id",
        Integer,
        ForeignKey("group.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    UniqueConstraint("user_id", "group_id"),
)

# Should not be needed (?)
administratorship = Table(
    "administratorship",
    db.Model.metadata,
    Column(
        "user_id",
        Integer,
        ForeignKey("user.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    Column(
        "group_id",
        Integer,
        ForeignKey("group.id", onupdate="CASCADE", ondelete="CASCADE"),
    ),
    UniqueConstraint("user_id", "group_id"),
)

_RANDOM_PASSWORD_CHARS = string.ascii_letters + string.digits + string.punctuation


def gen_random_password(length=15):
    rg = random.SystemRandom()
    return "".join(rg.choice(_RANDOM_PASSWORD_CHARS) for i in range(length))


class PasswordStrategy(metaclass=ABCMeta):
    @property
    @abstractmethod
    def name(self):
        """Strategy name."""

    @abstractmethod
    def authenticate(self, user, password):
        """Predicate to tell wether password match user's or not."""

    @abstractmethod
    def process(self, user, password):
        """Return a string to be stored as user password."""


class ClearPasswordStrategy(PasswordStrategy):
    """Don't encrypt at all.

    This strategy should not ever be used elsewhere than in tests. It's
    useful in tests since a hash like bcrypt is designed to be slow.
    """

    @property
    def name(self):
        return "clear"

    def authenticate(self, user, password):
        return user.password == password

    def process(self, user, password):
        if not isinstance(password, str):
            password = password.decode("utf-8")
        return password


class BcryptPasswordStrategy(PasswordStrategy):
    """Hash passwords using bcrypt."""

    @property
    def name(self):
        return "bcrypt"

    def authenticate(self, user: "User", password: str) -> bool:
        current_passwd = user.password
        # crypt work only on bytes, not str (Unicode)
        if isinstance(current_passwd, str):
            current_passwd = current_passwd.encode("utf-8")
        if isinstance(password, str):
            password = password.encode("utf-8")

        return bcrypt.hashpw(password, current_passwd) == current_passwd

    def process(self, user: "User", password: str) -> str:
        if isinstance(password, str):
            password = password.encode("utf-8")
        return bcrypt.hashpw(password, bcrypt.gensalt()).decode("utf-8")


class UserQuery(BaseQuery):
    def get_by_email(self, email):
        return self.filter_by(email=email).one()


class Principal(IdMixin, TimestampedMixin, Indexable):
    """A principal is either a User or a Group."""

    __index_to__ = (("name", ("name", "name_prefix", "text")),)

    def has_role(self, role, context=None):
        from abilian.services import get_service

        security = get_service("security")
        return security.has_role(self, role, context)


def set_entity_type(cls):
    """Decorator used to set the class' entity_type after the class has been
    declared.

    Actually, it works using __module__ during class declaration, but
    linters (Flake8, PyCharm) complain.
    """
    cls.entity_type = fqcn(cls)
    return cls


@set_entity_type
class User(Principal, UserMixin, db.Model):
    __tablename__ = "user"
    __editable__ = ["first_name", "last_name", "email", "password"]
    __exportable__ = __editable__ + ["created_at", "updated_at", "id"]

    __password_strategy__ = BcryptPasswordStrategy()

    query_class = UserQuery
    query: UserQuery

    # Basic information
    first_name = Column(UnicodeText, info=SEARCHABLE)
    last_name = Column(UnicodeText, info=SEARCHABLE)
    # Should we add gender, salutation ?

    # System information
    email = Column(UnicodeText, nullable=False)
    can_login = Column(Boolean, nullable=False, default=True)
    password = Column(UnicodeText, default="*", info={"audit_hide_content": True})

    photo = deferred(Column(LargeBinary))

    last_active = Column(DateTime, info=SYSTEM)
    locale = Column(sa_types.Locale, nullable=True, default=None)
    timezone = Column(sa_types.Timezone, nullable=True, default=None)

    __table_args__ = (UniqueConstraint("email"),)

    followers = relationship(
        "User",
        secondary=following,
        primaryjoin="User.id == following.c.follower_id",
        secondaryjoin="User.id == following.c.followee_id",
        backref="followees",
    )

    def __init__(self, password=None, **kwargs):
        Principal.__init__(self)
        UserMixin.__init__(self)
        db.Model.__init__(self, **kwargs)

        if self.can_login and password is not None:
            self.set_password(password)
            self._password = password

    def authenticate(self, password: str) -> bool:
        if self.password and self.password != "*":
            return self.__password_strategy__.authenticate(self, password)
        else:
            return False

    def set_password(self, password: str) -> None:
        """Encrypts and sets password."""
        # pyre-fixme[8]: Attribute has type `Column`; used as `str`.
        self.password = self.__password_strategy__.process(self, password)

    def follow(self, followee):
        if followee == self:
            raise Exception("User can't follow self")
        self.followees.append(followee)

    def unfollow(self, followee):
        if followee == self:
            raise Exception("User can't follow self")
        i = self.followees.index(followee)
        del self.followees[i]

    def join(self, group):
        self.groups.add(group)

    def leave(self, group):
        if group in self.groups:
            self.groups.remove(group)

    #
    # Boolean properties
    #
    def is_following(self, other):
        return other in self.followees

    def is_member_of(self, group):
        return self in group.members

    def is_admin_of(self, group):
        return self in group.admins

    @property
    def is_online(self):
        if self.last_active is None:
            return False
        return datetime.utcnow() - self.last_active <= timedelta(minutes=1)

    #
    # Other properties
    #
    @property
    def name(self) -> str:
        name = "{first_name} {last_name}".format(
            first_name=self.first_name or "", last_name=self.last_name or ""
        )
        return name.strip() or "Unknown"

    @property
    def short_name(self):
        first_name = self.first_name or ""
        last_name = self.last_name[0:1] + "." if self.last_name else ""
        name = f"{first_name} {last_name}"
        return name.strip() or "Unknown"

    def __str__(self) -> str:
        return self.name

    def __repr__(self):
        cls = self.__class__
        return "<{mod}.{cls} id={id!r} email={email!r} at 0x{addr:x}>".format(
            mod=cls.__module__,
            cls=cls.__name__,
            id=self.id,
            email=self.email,
            addr=id(self),
        )


@listens_for(User, "mapper_configured", propagate=True)
def _add_user_indexes(mapper: Mapper, class_: type) -> None:
    # this is a functional index (indexes on a function result), we cannot define
    # it in __table_args__.
    #
    # see: https://groups.google.com/d/msg/sqlalchemy/CgSJUlelhGs/_Nj3f201hs4J
    idx = sa.schema.Index(
        "user_unique_lowercase_email", sa.sql.func.lower(class_.email), unique=True
    )
    idx.info["engines"] = ("postgresql",)


@set_entity_type
class Group(Principal, db.Model):
    __indexable__ = False
    __tablename__ = "group"
    __editable__ = ["name", "description"]
    __exportable__ = __editable__ + ["created_at", "updated_at", "id"]

    name = Column(UnicodeText, nullable=False, info=SEARCHABLE)
    description = Column(UnicodeText, info=SEARCHABLE)

    members = relationship(
        "User",
        collection_class=set,
        secondary=membership,
        backref=backref("groups", lazy="select", collection_class=set),
    )
    admins = relationship("User", collection_class=set, secondary=administratorship)

    photo = deferred(Column(LargeBinary))

    public = Column(Boolean, default=False, nullable=False)

    @hybrid_property
    def members_count(self):
        return len(self.members)

    @members_count.expression
    def members_count(cls):
        return (
            sa.sql.select([sa.sql.func.count(membership.c.user_id)])
            .where(membership.c.group_id == cls.id)
            .group_by(membership.c.group_id)
            .label("members_count")
        )


def create_root_user() -> User:
    user = User.query.get(0)
    if user is None:
        user = User(
            id=0, last_name="SYSTEM", email="system@example.com", can_login=False
        )
        db.session.add(user)
        db.session.commit()
    return user



""""""
import os
import sys

import pkg_resources
from flask import current_app, render_template

from abilian.web.admin import AdminPanel


class SysinfoPanel(AdminPanel):
    id = "sysinfo"
    label = "System information"
    icon = "hdd"

    def get(self) -> str:
        uname = os.popen("uname -a").read()
        python_version = sys.version.strip()

        packages = []

        for dist in pkg_resources.working_set:
            package = {
                "name": dist.project_name,
                "key": dist.key,
                # pyre-fixme[16]: `Distribution` has no attribute `has_version`.
                "version": dist.version if dist.has_version() else "Unknown version",
                "vcs": None,
            }

            # FIXME: broken by pip 10
            # location = text_type(Path(dist.location).resolve())
            # vcs_name = get_backend_name(location)
            #
            # if vcs_name:
            #     vc = vcs.get_backend(vcs_name)()
            #     try:
            #         url = vc.get_url(location)
            #     except pip.exceptions.InstallationError:
            #         url = 'None'
            #     try:
            #         revision = vc.get_revision(location)
            #     except pip.exceptions.InstallationError:
            #         revision = 'None'
            #     package['vcs'] = dict(
            #         name=vcs_name,
            #         url=url,
            #         revision=revision,
            #     )

            packages.append(package)
            packages.sort(key=lambda d: d.get("key"))

        config_values = [(k, repr(v)) for k, v in sorted(current_app.config.items())]

        ctx = {
            "python_version": python_version,
            "packages": packages,
            "uname": uname,
            "config_values": config_values,
        }
        return render_template("admin/sysinfo.html", **ctx)



from typing import Union

import sqlalchemy as sa
from flask import Flask
from flask.ctx import RequestContext
from flask.signals import request_started
from pytest import mark
from wtforms import Form, IntegerField, StringField

# Import for side-effects (monkey-patch)
# noinspection PyUnresolvedReferences
import abilian.web.forms  # noqa
from abilian.core.entities import Entity
from abilian.web.forms.widgets import EmailWidget, MainTableView, Panel, Row, \
    SingleView, linkify_url, text2html
from abilian.web.views import default_view


class WidgetTestModel(Entity):
    """Mock model."""

    __tablename__ = "widget_test_model"
    id = sa.Column(sa.Integer, primary_key=True)
    price = sa.Column(sa.Integer)
    email = sa.Column(sa.Text)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._display_value_called = False

    def display_value(self, attr: str) -> Union[int, str]:
        self._display_value_called = True
        return getattr(self, attr)


class DummyForm(Form):
    name = StringField("Nom du véhicule")
    price = IntegerField("Prix du véhicule")
    email = StringField("email", view_widget=EmailWidget())


def test_table_view(app: Flask, test_request_context: RequestContext) -> None:
    @default_view(app, WidgetTestModel)
    @app.route("/dummy_view/<object_id>")
    def dummy_view(object_id):
        pass

    request_started.send(app)  # needed for deferJS tag
    columns = [{"name": "name"}, {"name": "price"}]
    view = MainTableView(columns)

    model1 = WidgetTestModel(id=1, name="Renault Megane", price=10000)
    model2 = WidgetTestModel(id=2, name="Peugeot 308", price=12000)
    models = [model1, model2]
    # pyre-fixme[6]: Expected `List[Entity]` for 1st param but got
    #  `List[WidgetTestModel]`.
    res = view.render(models)

    assert model1._display_value_called
    assert model2._display_value_called
    assert "Renault Megane" in res
    assert "10000" in res


def test_single_view(test_request_context: RequestContext) -> None:
    model = WidgetTestModel(name="Renault Megane", price=10000, email="joe@example.com")
    panels = [Panel("main", Row("name"), Row("price"), Row("email"))]
    form = DummyForm(obj=model)
    view = SingleView(form, *panels, view={"can_edit": False, "can_delete": False})
    res = view.render(model)

    assert "Renault Megane" in res
    assert "10000" in res
    # 'mailto:' is created by EmailWidget
    assert "mailto:joe@example.com" in res


@mark.skip
def test_edit_view(app):
    with app.test_request_context():
        model = WidgetTestModel(name="Renault Megane", price=10000)
        panels = [Panel("main", Row("name"), Row("price"))]
        form = DummyForm(obj=model)
        view = SingleView(form, *panels)
        res = view.render_form()

        assert model._display_value_called
        assert "Renault Megane" in res
        assert "10000" in res


EXPECTED = (
    '<a href="http://example.com">example.com</a>'
    '&nbsp;<i class="fa fa-external-link"></i>'
)


def test_http() -> None:
    value = "http://example.com"
    result = linkify_url(value)
    assert result == EXPECTED


def test_no_http() -> None:
    value = "example.com"
    result = linkify_url(value)
    assert result == EXPECTED


def test1() -> None:
    result = text2html("a")
    assert result == "a"


def test2() -> None:
    result = text2html("a\nb")
    assert str(result) == "<p>a</p>\n<p>b</p>"


def test3() -> None:
    result = text2html("a\n\nb")
    assert str(result) == "<p>a</p>\n<p>b</p>"


def test4() -> None:
    result = text2html("a\n<a>toto</a>")
    assert str(result) == "<p>a</p>\n<p>&lt;a&gt;toto&lt;/a&gt;</p>"



"""Tag service.

Manages tags applied to taggable entities.

Cf. ATOM specs  (+ for instance this discussion:
  http://edward.oconnor.cx/2007/02/representing-tags-in-atom)
Cf. ICOM 3.6.12 (Tag) et 3.6.13 (TagApplication.

Warning: the Activity Steams spec has a notion of tagging, but it's slightly
different than what we are used to in content management:
"Indicates that the actor has associated the object with the target.
For example, if the actor specifies that a particular user appears in a photo.
The object is the user and the target is the photo."
"""
from sqlalchemy import Column, DateTime, ForeignKey, Integer, Unicode

from abilian.core.extensions import db
from abilian.core.models import IdMixin


class Tag(db.Model, IdMixin):
    """Persistent class representing a tag."""

    #: Unique, autogenerated id.
    id = Column(Integer, primary_key=True)

    #: The text associated to the tag
    term = Column(Unicode, nullable=False, unique=True)

    # Not sure we want the extra complexity, even though it's specified by Atom.
    # label = Column()

    # Also we might want: created_at, etc. Maybe a Tag should be an Entity?


class TagApplication(db.Model, IdMixin):
    """Persistent class representing a tag application on a taggable object."""

    #: Unique, autogenerated id.
    id = Column(Integer, primary_key=True)

    #: The actor that did the tagging.
    actor_id = Column(Integer, ForeignKey("user.id"))

    #: The entity (object) on which the tag has been applied.
    entity_id = Column(Integer)  # must point to a valid entity

    #: The time when the tag was applied.
    created_at = Column(DateTime, nullable=False)



"""Activity Service.

See: http://activitystrea.ms/specs/json/1.0/
See: http://activitystrea.ms/specs/atom/1.0/#activity
See: http://stackoverflow.com/questions/1443960/

TODO: Look wether other attributes from the spec need to be implemented.
"""
import logging
from datetime import datetime

from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.types import DateTime, Integer, String, Text

from abilian.core.entities import Entity, db
from abilian.core.models.subjects import User

__all__ = ["ActivityEntry"]

logger = logging.getLogger(__name__)


def _default_from(column):
    """Helper for default and onupdates parameters in a Column definitions.

    Returns a `context-sensitive default function
    <http://docs.sqlalchemy.org/en/rel_0_8/core/defaults.html#context-
    sensitive-default-functions>`_ to set value from another column.
    """

    def _default_value(context):
        return context.current_parameters[column]

    return _default_value


class ActivityEntry(db.Model):
    """Main table for all activities."""

    id = Column(Integer, primary_key=True)
    happened_at = Column(DateTime, default=datetime.utcnow)

    verb = Column(Text)

    actor_id = Column(Integer, ForeignKey(User.id))
    actor = relationship(User, foreign_keys=actor_id)

    object_type = Column(String(1000))
    object_id = Column(Integer, default=_default_from("_fk_object_id"))
    _fk_object_id = Column(Integer, ForeignKey(Entity.id, ondelete="SET NULL"))
    object = relationship(Entity, foreign_keys=_fk_object_id)

    target_type = Column(String(1000))
    target_id = Column(Integer, default=_default_from("_fk_target_id"))
    _fk_target_id = Column(Integer, ForeignKey(Entity.id, ondelete="SET NULL"))
    target = relationship(Entity, foreign_keys=_fk_target_id)

    def __repr__(self):
        tpl = "<{}.ActivityEntry id={} actor={} verb={} object={} target={}>"
        return tpl.format(
            self.__class__.__module__,
            self.id,
            repr(str(self.actor)),
            repr(self.verb),
            repr(str(self.object)),
            repr(str(self.target)),
        )



""""""
from flask import url_for


def test_home(client, db_session):
    response = client.get(url_for("admin.dashboard"))
    assert response.status_code == 200


def test_sysinfo(client, db_session):
    response = client.get(url_for("admin.sysinfo"))
    assert response.status_code == 200


def test_login_session(client, db_session):
    response = client.get(url_for("admin.login_sessions"))
    assert response.status_code == 200


def test_audit(client, db_session):
    response = client.get(url_for("admin.audit"))
    assert response.status_code == 200


def test_settings(client, db_session):
    response = client.get(url_for("admin.settings"))
    assert response.status_code == 200



""""""
from datetime import datetime
from typing import Any

from flask_sqlalchemy import BaseQuery
from sqlalchemy.schema import Column
from sqlalchemy.types import DateTime, Integer
from whoosh.fields import ID

from abilian.core.extensions import db
from abilian.core.util import fqcn


#: Base Model class.
class Model(db.Model):
    __abstract__ = True
    query: BaseQuery


class Info(dict):
    def __init__(self, **kw: Any) -> None:
        for k, v in kw.items():
            self[k] = v

    def copy(self) -> "Info":
        # dict.copy would return an instance of dict
        # pyre-fixme[32]: Keyword argument must be a mapping with string keys.
        return self.__class__(**self)

    def __add__(self, other: "Info") -> "Info":
        d = self.copy()
        d.update(other)
        return d

    __or__ = __add__


EDITABLE = Info(editable=True)
NOT_EDITABLE = Info(editable=False)
AUDITABLE = Info(auditable=True)
AUDITABLE_HIDDEN = Info(auditable=True, audit_hide_content=True)
NOT_AUDITABLE = Info(auditable=False)
SEARCHABLE = Info(searchable=True)
NOT_SEARCHABLE = Info(searchable=False)
EXPORTABLE = Info(exportable=True)
NOT_EXPORTABLE = Info(exportable=False)

#: SYSTEM properties are properties defined by the system
#: and not supposed to be changed manually.
SYSTEM = Info(editable=False, auditable=False)


class IdMixin:
    id = Column(Integer, primary_key=True, info=SYSTEM | SEARCHABLE)


class Indexable:
    """Mixin with sensible defaults for indexable objects."""

    __indexable__ = True
    __index_to__ = (
        ("object_key", (("object_key", ID(stored=True, unique=True)),)),
        ("object_type", (("object_type", ID(stored=True)),)),
    )

    @classmethod
    def _object_type(cls) -> str:
        return fqcn(cls)

    @property
    def object_type(self) -> str:
        return self._object_type()

    @property
    def object_key(self) -> str:
        # pyre-fixme[16]: `Indexable` has no attribute `id`.
        return f"{self.object_type}:{self.id}"


class TimestampedMixin:
    #: creation date
    created_at = Column(DateTime, default=datetime.utcnow, info=SYSTEM | SEARCHABLE)
    #: last modification date
    updated_at = Column(
        DateTime,
        default=datetime.utcnow,
        onupdate=datetime.utcnow,
        info=SYSTEM | SEARCHABLE,
    )
    deleted_at = Column(DateTime, default=None, info=SYSTEM)



""""""
from flask import current_app
from flask_debugtoolbar.panels import DebugPanel

from abilian.core.entities import Entity
from abilian.core.models.subjects import Group
from abilian.i18n import _
from abilian.services import get_service
from abilian.web.action import actions

from .models import Anonymous


class SecurityInfoDebugPanel(DebugPanel):
    """A panel to display current roles and permissions for "current"
    object."""

    name = "SecurityInfo"

    @property
    def current_obj(self):
        return actions.context.get("object")

    @property
    def has_content(self):
        obj = self.current_obj
        return obj is not None and isinstance(obj, Entity) and obj.id is not None

    def nav_title(self):
        return _("Security Info")

    def nav_subtitle(self):
        """Subtitle showing until title in toolbar."""
        obj = self.current_obj
        if not obj:
            return _("No current object")

        try:
            return f"{obj.__class__.__name__}(id={obj.id})"
        except Exception:
            return ""

    def title(self):
        return self.nav_title()

    def url(self):
        return ""

    def content(self):
        obj = self.current_obj
        security = get_service("security")
        context = self.context.copy()

        context["permissions"] = security.get_permissions_assignments(obj=obj)
        context["roles"] = roles = {}

        for principal, r in security.get_role_assignements(obj=obj):
            if r not in roles:
                roles[r] = {"anonymous": False, "users": set(), "groups": set()}

            info = roles[r]
            if principal is Anonymous:
                info["anonymous"] = True
            elif isinstance(principal, Group):
                info["groups"].add(principal)
            else:
                info["users"].add(principal)

        for r in roles:
            info = roles[r]
            info["groups"] = [
                "{g} (id={g.id})".format(g=g)
                for g in sorted(info["groups"], key=lambda g: g.name)
            ]
            users = sorted(
                info["users"], key=lambda u: (u.last_name.lower(), u.first_name.lower())
            )
            info["users"] = [
                '{u} (id={u.id}, email="{u.email}")'.format(u=u) for u in users
            ]

        jinja_env = current_app.jinja_env
        jinja_env.filters.update(self.jinja_env.filters)
        template = jinja_env.get_or_select_template(
            "debug_panels/security_info_panel.html"
        )

        return template.render(context)



"""Conversion service.

Hardcoded to manage only conversion to PDF, to text and to image series.

Includes result caching (on filesystem).

Assumes poppler-utils and LibreOffice are installed.

TODO: rename Converter into ConversionService ?
"""
import logging
import shutil
import subprocess
from io import BytesIO
from pathlib import Path

from flask import Flask
from PIL import Image
from PIL.ExifTags import TAGS

from .cache import Cache
from .util import make_temp_file

logger = logging.getLogger(__name__)

TMP_DIR = "tmp"
CACHE_DIR = "cache"


class ConversionError(Exception):
    pass


class HandlerNotFound(ConversionError):
    pass


class Converter:
    def __init__(self):
        self.handlers = []
        self.cache = Cache()

    def init_app(self, app: Flask) -> None:
        self.init_work_dirs(
            cache_dir=Path(app.instance_path, CACHE_DIR),
            tmp_dir=Path(app.instance_path, TMP_DIR),
        )

        app.extensions["conversion"] = self

        for handler in self.handlers:
            handler.init_app(app)

    def init_work_dirs(self, cache_dir: Path, tmp_dir: Path):
        self.tmp_dir = tmp_dir
        self.cache_dir = cache_dir
        self.cache.cache_dir = self.cache_dir

        if not self.tmp_dir.exists():
            self.tmp_dir.mkdir()
        if not self.cache_dir.exists():
            self.cache_dir.mkdir()

    def clear(self) -> None:
        self.cache.clear()
        # pyre-fixme[16]: `Converter` has no attribute `tmp_dir`.
        # pyre-fixme[16]: `Converter` has no attribute `cache_dir`.
        for d in (self.tmp_dir, self.cache_dir):
            shutil.rmtree(bytes(d))
            d.mkdir()

    def register_handler(self, handler):
        self.handlers.append(handler)

    # TODO: refactor, pass a "File" or "Document" or "Blob" object
    def to_pdf(self, digest: str, blob: bytes, mime_type: str) -> bytes:
        cache_key = ("pdf", digest)
        pdf = self.cache.get(cache_key)
        if pdf:
            return pdf

        for handler in self.handlers:
            if handler.accept(mime_type, "application/pdf"):
                pdf = handler.convert(blob)
                self.cache[cache_key] = pdf
                return pdf
        raise HandlerNotFound(f"No handler found to convert from {mime_type} to PDF")

    def to_text(self, digest: str, blob: bytes, mime_type: str) -> str:
        """Convert a file to plain text.

        Useful for full-text indexing. Returns a Unicode string.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = ("txt", digest)

        text = self.cache.get(cache_key)
        if text:
            return text

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "text/plain"):
                text = handler.convert(blob)
                self.cache[cache_key] = text
                return text

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "text/plain"):
                text = handler.convert(pdf)
                self.cache[cache_key] = text
                return text

        raise HandlerNotFound(f"No handler found to convert from {mime_type} to text")

    def has_image(self, digest, mime_type, index, size=500):
        """Tell if there is a preview image."""
        cache_key = (f"img:{index}:{size}", digest)
        return mime_type.startswith("image/") or cache_key in self.cache

    def get_image(self, digest, blob, mime_type, index, size=500):
        """Return an image for the given content, only if it already exists in
        the image cache."""
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = (f"img:{index}:{size}", digest)
        return self.cache.get(cache_key)

    def to_image(
        self, digest: str, blob: bytes, mime_type: str, index: int, size: int = 500
    ) -> bytes:
        """Convert a file to a list of images.

        Returns image at the given index.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return b""

        cache_key = (f"img:{index}:{size}", digest)
        converted = self.cache.get(cache_key)
        if converted:
            return converted

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "image/jpeg"):
                converted_images = handler.convert(blob, size=size)
                for i in range(0, len(converted_images)):
                    converted = converted_images[i]
                    cache_key = (f"img:{i}:{size}", digest)
                    self.cache[cache_key] = converted
                return converted_images[index]

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "image/jpeg"):
                converted_images = handler.convert(pdf, size=size)
                for i in range(0, len(converted_images)):
                    converted = converted_images[i]
                    cache_key = (f"img:{i}:{size}", digest)
                    self.cache[cache_key] = converted
                return converted_images[index]

        raise HandlerNotFound(f"No handler found to convert from {mime_type} to image")

    def get_metadata(self, digest, content, mime_type):
        """Get a dictionary representing the metadata embedded in the given
        content."""

        # XXX: ad-hoc for now, refactor later
        if mime_type.startswith("image/"):
            img = Image.open(BytesIO(content))
            ret = {}
            if not hasattr(img, "_getexif"):
                return {}
            info = img._getexif()
            if not info:
                return {}
            for tag, value in info.items():
                decoded = TAGS.get(tag, tag)
                ret["EXIF:" + str(decoded)] = value
            return ret

        else:
            if mime_type != "application/pdf":
                content = self.to_pdf(digest, content, mime_type)

            with make_temp_file(content) as in_fn:
                try:
                    output = subprocess.check_output(["pdfinfo", in_fn])
                except OSError:
                    logger.error("Conversion failed, probably pdfinfo is not installed")
                    raise

            ret = {}
            for line in output.split(b"\n"):
                if b":" in line:
                    key, value = line.strip().split(b":", 1)
                    key = str(key)
                    ret["PDF:" + key] = str(value.strip(), errors="replace")

            return ret



""""""
from datetime import datetime

from flask import request
from flask_login import current_user
from flask_sqlalchemy import BaseQuery
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import relation

from abilian.core.extensions import db
from abilian.core.models.subjects import User

_MARK = object()


class LoginSessionQuery(BaseQuery):
    def get_active_for(
        self, user: User, user_agent: object = _MARK, ip_address: object = _MARK
    ) -> "LoginSession":
        """Return last known session for given user.

        :param user: user session
        :type user: `abilian.core.models.subjects.User`

        :param user_agent: *exact* user agent string to lookup, or `None` to have
        user_agent extracted from request object. If not provided at all, no
        filtering on user_agent.
        :type user_agent: string or None, or absent

        :param ip_address: client IP, or `None` to have ip_address extracted from
        request object (requires header 'X-Forwarded-For'). If not provided at
        all, no filtering on ip_address.
        :type ip_address: string or None, or absent

        :rtype: `LoginSession` or `None` if no session is found.
        """
        conditions = [LoginSession.user == user]

        if user_agent is not _MARK:
            if user_agent is None:
                user_agent = request.environ.get("HTTP_USER_AGENT", "")
            conditions.append(LoginSession.user_agent == user_agent)

        if ip_address is not _MARK:
            if ip_address is None:
                ip_addresses = request.headers.getlist("X-Forwarded-For")
                ip_address = ip_addresses[0] if ip_addresses else request.remote_addr
            conditions.append(LoginSession.ip_address == ip_address)

        session = (
            LoginSession.query.filter(*conditions)
            .order_by(LoginSession.id.desc())
            .first()
        )
        return session


class LoginSession(db.Model):

    __tablename__ = "login_session"
    query_class = LoginSessionQuery
    query: LoginSessionQuery

    id = Column(Integer, primary_key=True, autoincrement=True)

    started_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    last_active_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    ended_at = Column(DateTime)

    user_id = Column(ForeignKey(User.id), nullable=False)
    user = relation(User)

    #: IP addres, should work with both IPv4 and IPv6.
    # FIXME: define a custom type to use postgres inet when on postgres
    # http://www.postgresql.org/docs/9.1/static/datatype-net-types.html
    ip_address = Column(String(39), default="", nullable=False)

    #: Limited (somewhat artbitrarily) to 512 characters.
    user_agent = Column(String(512), default="", nullable=False)

    @staticmethod
    def new() -> "LoginSession":
        user_agent = request.environ.get("HTTP_USER_AGENT", "")
        forwared_for = request.headers.getlist("X-Forwarded-For")
        if not forwared_for:
            ip_address = request.remote_addr
        else:
            ip_address = forwared_for[0]
        session = LoginSession(
            user=current_user, user_agent=user_agent, ip_address=ip_address
        )
        return session



"""audit Service: logs modifications to audited objects.

TODO: In the future, we may decide to:

- Make Models that have the __auditable__ property (set to True) auditable.
- Make Entities that have the __auditable__ property set to False not auditable.
"""
import logging
from datetime import datetime
from inspect import isclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

import sqlalchemy as sa
from flask import current_app, g
from sqlalchemy import event, extract
from sqlalchemy.orm import Query, Session
from sqlalchemy.orm.attributes import NEVER_SET, InstrumentedAttribute
from sqlalchemy.orm.unitofwork import UOWTransaction

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.services import Service, ServiceState

from .models import CREATION, DELETION, RELATED, UPDATE, AuditEntry, Changes

if TYPE_CHECKING:
    from abilian.app import Application

log = logging.getLogger(__name__)


class AuditableMeta:
    backref_attr: Optional[str] = None
    audited_attrs: Set[InstrumentedAttribute]
    collection_attrs: Set[InstrumentedAttribute]
    enduser_ids: List[str]

    def __init__(
        self,
        name: Optional[str] = None,
        id_attr: Optional[str] = None,
        related: bool = False,
    ) -> None:
        self.name = name
        self.id_attr = id_attr
        self.related = related
        self.audited_attrs = set()
        self.collection_attrs = set()
        self.enduser_ids = []


class AuditServiceState(ServiceState):

    all_model_classes = None
    model_class_names = None

    # set to True when creating audit entries, to avoid examining a session full
    # of audit entries
    creating_entries = False

    def __init__(self, service: "AuditService", *args: Any, **kwargs: Any) -> None:
        super().__init__(service, *args, **kwargs)
        self.all_model_classes: Set[type] = set()
        self.model_class_names: Dict[str, type] = {}


class AuditService(Service):
    name = "audit"
    AppStateClass = AuditServiceState

    _listening = False

    def init_app(self, app: "Application") -> None:
        super().init_app(app)

        if not self._listening:
            event.listen(Session, "after_flush", self.create_audit_entries)
            self._listening = True

    def start(self, ignore_state: bool = False) -> None:
        super().start(ignore_state)
        self.register_classes()

    @staticmethod
    def is_auditable(model_or_class):
        if hasattr(model_or_class, "__auditable_entity__"):
            return True

        if isclass(model_or_class):
            return issubclass(model_or_class, Entity)
        else:
            return isinstance(model_or_class, Entity)

    def register_classes(self) -> None:
        state = self.app_state
        BaseModel = db.Model
        all_models = (
            cls
            for cls in BaseModel._decl_class_registry.values()
            if isclass(cls) and self.is_auditable(cls)
        )
        for cls in all_models:
            self.register_class(cls, app_state=state)

    def register_class(self, entity_class, app_state=None):
        if not hasattr(entity_class, "__table__"):
            return

        state = app_state if app_state is not None else self.app_state
        if entity_class in state.all_model_classes:
            return

        state.all_model_classes.add(entity_class)
        self.setup_auditable_entity(entity_class)

        assert entity_class.__name__ not in state.model_class_names
        state.model_class_names[entity_class.__name__] = entity_class

        mapper = sa.orm.class_mapper(entity_class)
        for column in mapper.columns:
            props = mapper.get_property_by_column(column)
            attr = getattr(entity_class, props.key)
            info = column.info

            if info.get("auditable", True):
                entity_class.__auditable__.audited_attrs.add(attr)
                event.listen(attr, "set", self.set_attribute, active_history=True)

        for relation in mapper.relationships:
            if relation.direction is not sa.orm.interfaces.MANYTOMANY:
                continue
            attr = getattr(entity_class, relation.key)
            entity_class.__auditable__.collection_attrs.add(attr)
            event.listen(attr, "append", self.collection_append, active_history=True)
            event.listen(attr, "remove", self.collection_remove, active_history=True)

    def setup_auditable_entity(self, entity_class):
        meta = AuditableMeta(entity_class.__name__, "id")
        entity_class.__auditable__ = meta

        if not hasattr(entity_class, "__auditable_entity__"):
            return

        related_attr, backref_attr, enduser_ids = entity_class.__auditable_entity__
        related_path = related_attr.split(".")
        inferred_backref = []
        mapper = sa.orm.class_mapper(entity_class)
        for attr in related_path:
            relation = mapper.relationships.get(attr)
            if not relation:
                raise ValueError(
                    'Invalid relation: "{}", invalid attribute is "{}"'
                    "".format(related_attr, attr)
                )

            mapper = relation.mapper
            if inferred_backref is not None:
                try:
                    inferred_backref.append(relation.back_populates)
                except AttributeError:
                    inferred_backref = None

        try:
            meta.name = mapper.entity.__name__
        except AttributeError:
            return

        if not backref_attr:
            if inferred_backref is not None:
                backref_attr = ".".join(inferred_backref)
            else:
                raise ValueError(
                    "Audit setup class<{cls}: Could not guess backref name"
                    ' of relationship "{related_attr}", please use tuple annotation '
                    "on __auditable_entity__".format(
                        cls=entity_class.__name__, related_attr=related_attr
                    )
                )

        meta.related = related_path
        meta.backref_attr = backref_attr
        meta.enduser_ids = [attr_name.split(".") for attr_name in enduser_ids]

    def _get_changes_for(self, entity):
        changes = getattr(entity, "__changes__", None)
        if not changes:
            changes = entity.__changes__ = Changes()
        return changes

    def set_attribute(self, entity, new_value, old_value, initiator):
        attr_name = initiator.key
        if old_value == new_value:
            return

        # We don't log a few trivial cases so as not to overflow the audit log.
        if not old_value and not new_value:
            return

        changes = self._get_changes_for(entity)
        if attr_name in changes.columns:
            old_value = changes.columns[attr_name][0]

        # Hide content if needed (like password columns)
        # FIXME: we can only handle the simplest case: 1 attribute => 1 column
        columns = initiator.parent_token.columns
        if len(columns) == 1 and columns[0].info.get("audit_hide_content"):
            old_value = new_value = "******"

        old_value = format_large_value(old_value)
        new_value = format_large_value(new_value)
        changes.set_column_changes(attr_name, old_value, new_value)

    def collection_append(self, entity, value, initiator):
        changes = self._get_changes_for(entity)
        changes.collection_append(initiator.key, value)

    def collection_remove(self, entity, value, initiator):
        changes = self._get_changes_for(entity)
        changes.collection_remove(initiator.key, value)

    def create_audit_entries(
        self, session: Session, flush_context: UOWTransaction
    ) -> None:
        if not self.running or self.app_state.creating_entries:
            return

        self.app_state.creating_entries = True
        try:
            # if an error happens during audit creation it should not break the rest of
            # the application, and db session should be left clean. Only the developper
            # (and raven/sentry/whatever) should know
            entries = []
            for identity_set, op in (
                (session.new, CREATION),
                (session.deleted, DELETION),
                (session.dirty, UPDATE),
            ):
                for model in identity_set:
                    try:
                        entry = self.log(session, model, op)
                        if entry:
                            entries.append(entry)
                    except Exception:
                        if current_app.debug or current_app.testing:
                            raise
                        log.error("Exception during entry creation", exc_info=True)

                session.add_all(entries)
        finally:
            self.app_state.creating_entries = False

    def log(self, session, model, op_type):
        if not self.is_auditable(model):
            return

        entity = model
        try:
            user_id = g.user.id
        except Exception:
            user_id = 0

        meta = model.__auditable__
        if meta.related:
            op_type |= RELATED
            entity = model
            for attr in meta.related:
                entity = getattr(entity, attr)
                if entity is None:
                    return

        entry = AuditEntry(type=op_type, user_id=user_id)
        if op_type != DELETION:
            # DELETION|RELATED: deletion of a related model is ok: entity is still
            # here
            entry.entity = entity

        entry.entity_id = entity.id
        entry.entity_type = entity.entity_type

        entity_name = ""
        for attr_name in ("name", "path", "__path_before_delete"):
            if hasattr(entity, attr_name):
                entity_name = getattr(entity, attr_name)
        entry.entity_name = entity_name

        changes = Changes()
        op = entry.op
        if op == CREATION:
            for instrumented_attr in meta.audited_attrs:
                value = getattr(model, instrumented_attr.key)
                self.set_attribute(model, value, NEVER_SET, instrumented_attr.impl)

            for instrumented_attr in meta.collection_attrs:
                for obj in getattr(model, instrumented_attr.key):
                    self.collection_append(model, obj, instrumented_attr.impl)

            changes = getattr(model, "__changes__", changes)
        elif op == UPDATE:
            changes = getattr(model, "__changes__", changes)
            if not changes:
                return

        if hasattr(model, "__changes__"):
            del model.__changes__

        if meta.related:
            enduser_ids = []
            for path in meta.enduser_ids:
                item = model
                for attr in path:
                    item = getattr(item, attr)
                enduser_ids.append(str(item))

            related_name = "{} {}".format(meta.backref_attr, " ".join(enduser_ids))
            related_changes = changes
            log.debug("related changes: %s", repr(related_changes))
            changes = Changes()
            changes.set_related_changes(related_name, related_changes)

        entry.changes = changes
        return entry

    def entries_for(self, entity, limit=None):
        query = AuditEntry.query.filter(AuditEntry.entity == entity).order_by(
            AuditEntry.happened_at.desc()
        )

        if limit is not None:
            query = query.limit(limit)

        return query.all()


audit_service = AuditService()


def format_large_value(value):
    try:
        if len(value) > 1000:
            return "<<large value>>"
    except TypeError:
        # object of type '...' has no len()
        pass
    return value


def get_model_changes(
    entity_type: str,
    year: Optional[int] = None,
    month: Optional[int] = None,
    day: Optional[int] = None,
    hour: Optional[int] = None,
    since: Optional[datetime] = None,
) -> Query:
    """Get models modified at the given date with the Audit service.

    :param entity_type: string like "extranet_medicen.apps.crm.models.Compte".
      Beware the typo, there won't be a warning message.
    :param since: datetime
    :param year: int
    :param month: int
    :param day: int
    :param hour: int

    :returns: a query object
    """
    query = AuditEntry.query

    if since:
        query = query.filter(AuditEntry.happened_at >= since)

    if year:
        query = query.filter(extract("year", AuditEntry.happened_at) == year)
    if month:
        query = query.filter(extract("month", AuditEntry.happened_at) == month)
    if day:
        query = query.filter(extract("day", AuditEntry.happened_at) == day)
    if hour:
        query = query.filter(extract("hour", AuditEntry.happened_at) == hour)

    query = query.filter(AuditEntry.entity_type.like(entity_type)).order_by(
        AuditEntry.happened_at
    )

    return query


def get_columns_diff(changes):
    """Add the changed columns as a diff attribute.

    - changes: a list of changes (get_model_changes query.all())

    Return: the same list, to which elements we added a "diff"
    attribute containing the changed columns. Diff defaults to [].
    """
    for change in changes:
        change.diff = []
        elt_changes = change.get_changes()
        if elt_changes:
            change.diff = elt_changes.columns

    return changes



""""""
import logging
from typing import Any, Callable, Dict, List, Optional, Union

from flask import Blueprint, Flask, g
from flask.helpers import _endpoint_from_view_func
from flask_login import current_user
from werkzeug.exceptions import Forbidden
from werkzeug.utils import import_string

from abilian.core.util import unwrap
from abilian.i18n import _l
from abilian.services.security import Admin as AdminRole
from abilian.services.security import security
from abilian.web.action import Endpoint, actions
from abilian.web.nav import BreadcrumbItem, NavGroup, NavItem

from .panel import AdminPanel

logger = logging.getLogger(__name__)

_BP_PREFIX = "admin"


class Admin:
    """Flask extension for an admin interface with pluggable admin panels.

    Note: this is quite different that a Django-style admin interface.
    """

    def __init__(self, *panels: Any, **kwargs: Any) -> None:
        self.app = None
        self.panels: List[AdminPanel] = []
        self._panels_endpoints: Dict[str, AdminPanel] = {}
        self.nav_paths: Dict[str, str] = {}
        self.breadcrumb_items: Dict[AdminPanel, BreadcrumbItem] = {}
        self.setup_blueprint()

        def condition(context: Dict[str, bool]) -> bool:
            return not current_user.is_anonymous and security.has_role(
                # pyre-fixme[6]: Expected `Principal` for 1st param but got
                #  `LocalProxy`.
                current_user,
                AdminRole,
            )

        self.nav_root = NavGroup(
            "admin", "root", title=_l("Admin"), endpoint=None, condition=condition
        )

        for panel in panels:
            self.register_panel(panel)

        app = kwargs.pop("app", None)
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        panels = app.config.get("ADMIN_PANELS", ())

        # resolve fully qualified name into an AdminPanel object
        for fqn in panels:
            panel_class = import_string(fqn, silent=True)
            if panel_class is None:
                logger.warning('Could not import panel: "%s"', fqn)
                continue
            if not issubclass(panel_class, AdminPanel):
                logger.error(
                    '"%s" is not a %s.AdminPanel, skipping', fqn, AdminPanel.__module__
                )
                continue

            self.register_panel(panel_class())
            logger.debug('Registered panel "%s"', fqn)

        if not self.panels:

            @self.blueprint.route("", endpoint="no_panel")
            def no_panels_view():
                return "No panels registered"

            self.nav_root.endpoint = "admin.no_panel"
        else:
            self.nav_root.endpoint = self.nav_root.items[0].endpoint

        self.root_breadcrumb_item = BreadcrumbItem(
            label=self.nav_root.title, url=self.nav_root.endpoint
        )

        app.register_blueprint(self.blueprint)

        with app.app_context():
            actions.register(self.nav_root, *self.nav_root.items)

        self.app = app
        app.extensions["admin"] = self

    def register_panel(self, panel: Any) -> None:
        if self.app:
            raise ValueError(
                "Extension already initialized for app, cannot add more" " panel"
            )

        self.panels.append(panel)
        panel.admin = self
        rule = "/" + panel.id
        endpoint = nav_id = panel.id
        abs_endpoint = f"admin.{endpoint}"

        if hasattr(panel, "get"):
            self.blueprint.add_url_rule(rule, endpoint, panel.get)
            self._panels_endpoints[abs_endpoint] = panel
        if hasattr(panel, "post"):
            post_endpoint = endpoint + "_post"
            self.blueprint.add_url_rule(
                rule, post_endpoint, panel.post, methods=["POST"]
            )
            self._panels_endpoints["admin." + post_endpoint] = panel

        panel.install_additional_rules(
            self.get_panel_url_rule_adder(panel, rule, endpoint)
        )

        nav = NavItem(
            "admin:panel",
            nav_id,
            title=panel.label,
            icon=panel.icon,
            endpoint=abs_endpoint,
        )
        self.nav_root.append(nav)
        self.nav_paths[abs_endpoint] = nav.path
        self.breadcrumb_items[panel] = BreadcrumbItem(
            label=panel.label, icon=panel.icon, url=Endpoint(abs_endpoint)
        )

    def get_panel_url_rule_adder(
        self, panel: Any, base_url: str, base_endpoint: str
    ) -> Callable:
        extension = self

        def add_url_rule(
            rule: str,
            endpoint: Optional[Any] = None,
            view_func: Optional[Callable] = None,
            **kwargs: Any,
        ) -> None:
            if not rule:
                # '' is already used for panel get/post
                raise ValueError("Invalid additional url rule: {}".format(repr(rule)))

            if endpoint is None:
                # pyre-fixme[16]: Module `helpers` has no attribute
                #  `_endpoint_from_view_func`.
                endpoint = _endpoint_from_view_func(view_func)

            if not endpoint.startswith(base_endpoint):
                endpoint = base_endpoint + "_" + endpoint

            extension._panels_endpoints["admin." + endpoint] = panel
            self.blueprint.add_url_rule(
                base_url + rule, endpoint=endpoint, view_func=view_func, **kwargs
            )

        return add_url_rule

    def setup_blueprint(self) -> None:
        self.blueprint = Blueprint(
            "admin", __name__, template_folder="templates", url_prefix="/" + _BP_PREFIX
        )

        self.blueprint.url_value_preprocessor(self.build_breadcrumbs)
        self.blueprint.url_value_preprocessor(self.panel_preprocess_value)

        @self.blueprint.before_request
        def check_security() -> None:
            user = unwrap(current_user)
            if not security.has_role(user, "admin"):
                raise Forbidden()

    def panel_preprocess_value(self, endpoint, view_args):
        panel = self._panels_endpoints.get(endpoint)
        if panel is not None:
            panel.url_value_preprocess(endpoint, view_args)

    def build_breadcrumbs(self, endpoint, view_args):
        g.breadcrumb.append(self.root_breadcrumb_item)
        g.nav["active"] = self.nav_paths.get(endpoint, self.nav_root.path)
        panel = self._panels_endpoints.get(endpoint)
        if panel:
            endpoint_bc = self.breadcrumb_items.get(panel)
            if endpoint_bc:
                g.breadcrumb.append(endpoint_bc)



""""""
from typing import Dict, List

import bleach
from wtforms.fields import BooleanField, IntegerField, StringField
from wtforms.widgets import HiddenInput

from abilian.i18n import _l
from abilian.web.forms import ModelForm
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required

ALLOWED_TAGS = ["b", "i", "del", "s", "u", "small", "strong", "em"]
ALLOWED_ATTRIBUTES = {}  # type: Dict[str, List[str]]


class EditForm(ModelForm):
    label = StringField(
        _l("Label"),
        description=_l("allowed tags: %(tags)s", tags=", ".join(ALLOWED_TAGS)),
        filters=(strip,),
        validators=[required()],
    )
    default = BooleanField(_l("Default"), default=False)
    active = BooleanField(_l("Active"), default=True)

    def validate_label(self, field):
        field.data = bleach.clean(
            field.data, tags=ALLOWED_TAGS, attributes=ALLOWED_ATTRIBUTES, strip=True
        )


class ListEditForm(EditForm):
    id = IntegerField(widget=HiddenInput())
    position = IntegerField(widget=HiddenInput())



""""""
from wtforms.fields import BooleanField, StringField

from abilian.i18n import _l
from abilian.services.security.models import Role
from abilian.web.forms import Form, fields, widgets
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required


class GroupAdminForm(Form):
    name = StringField(_l("Name"), filters=(strip,), validators=[required()])
    description = StringField(_l("Description"), filters=(strip,))

    public = BooleanField(_l("Public"), widget=widgets.BooleanWidget(on_off_mode=True))

    roles = fields.Select2MultipleField(
        _l("Roles"),
        choices=lambda: [(r.name, r.label) for r in Role.assignable_roles()],
    )



""""""
import sys
from typing import List, Text

from blinker import Namespace, Signal
from flask import current_app, g
from flask_debugtoolbar.panels import DebugPanel

from abilian.web.action import actions


class ActionDebugPanel(DebugPanel):
    name = "Actions"

    user_enable = True
    has_content = True

    def nav_title(self):
        return "Actions"

    def title(self):
        return "Actions"

    def url(self):
        return ""

    def content(self):
        # type: () -> Text
        actions_for_template = []

        for category in actions.actions().keys():
            available_actions = actions.for_category(category)
            for action in available_actions:
                d = {
                    "category": action.category,
                    "title": action.title,
                    "class": action.__class__.__name__,
                }
                try:
                    d["endpoint"] = str(action.endpoint)
                except Exception:
                    d["endpoint"] = "<Exception>"
                try:
                    d["url"] = str(action.url(g.action_context))
                except Exception:
                    d["url"] = "<Exception>"
                actions_for_template.append(d)

        actions_for_template.sort(key=lambda x: (x["category"], x["title"]))

        ctx = {"actions": actions_for_template}

        jinja_env = current_app.jinja_env
        jinja_env.filters.update(self.jinja_env.filters)
        template = jinja_env.get_or_select_template("debug_panels/actions_panel.html")
        return template.render(ctx)


class SignalsDebugPanel(DebugPanel):
    name = "Signals"

    user_enable = True
    has_content = True

    events = []  # type: List[dict]

    def nav_title(self):
        return "Signals"

    def title(self):
        return "Signals"

    def url(self):
        return ""

    def content(self):
        # type: () -> Text
        module_names = sorted(sys.modules.keys())

        signals = []

        for module_name in module_names:
            module = sys.modules[module_name]
            if not module:
                continue
            module_vars = vars(module)
            for var_name, var in module_vars.items():
                if not isinstance(var, Namespace):
                    continue

                ns = var
                ns_name = var_name
                for signal_name, signal in ns.items():
                    receivers = [str(r) for r in signal.receivers.values()]
                    d = {
                        "module_name": module_name,
                        "ns_name": ns_name,
                        "signal_name": signal_name,
                        "signal": signal,
                        "receivers": receivers,
                    }
                    signals.append(d)

        # pyre-fixme[18]: Global name `d` is undefined.
        signals.sort(key=lambda x: (d["module_name"], d["ns_name"], d["signal_name"]))

        ctx = {"signals": signals, "events": self.events}

        jinja_env = current_app.jinja_env
        jinja_env.filters.update(self.jinja_env.filters)
        template = jinja_env.get_or_select_template("debug_panels/signals_panel.html")
        return template.render(ctx)

    def process_request(self, request):
        self.events = []

        if getattr(Signal.send, "__wrapped", False):
            return

        orig_send = Signal.send
        events = self.events

        def wrapped_send(self, *sender, **kwargs):
            d = {"signal_name": self.name, "sender": str(sender[0]), "args": kwargs}
            events.append(d)
            return orig_send(self, *sender, **kwargs)

        wrapped_send.__wrapped = True
        Signal.send = wrapped_send



import uuid

from sqlalchemy import Column
from sqlalchemy.orm import Query, Session

from abilian.core.entities import Entity
from abilian.core.sqlalchemy import UUID, JSONDict, JSONList


class DummyModel2(Entity):
    list_attr = Column(JSONList())
    dict_attr = Column(JSONDict())
    uuid = Column(UUID())

    query: Query


def test_list_attribute(session: Session) -> None:
    model = DummyModel2(list_attr=[1, 2, 3])
    session.add(model)
    session.commit()
    model_id = model.id
    # pyre-fixme[16]: `Session` has no attribute `remove`.
    session.remove()
    model2 = DummyModel2.query.get(model_id)
    assert model2.list_attr == [1, 2, 3]

    model2.list_attr.append(4)
    assert model2.list_attr == [1, 2, 3, 4]

    del model2.list_attr[0]
    assert model2.list_attr == [2, 3, 4]


def test_dict_attribute(session: Session) -> None:
    model = DummyModel2(dict_attr={"a": 3, "b": 4})
    session.add(model)
    session.commit()
    model_id = model.id
    # pyre-fixme[16]: `Session` has no attribute `remove`.
    session.remove()
    model2 = DummyModel2.query.get(model_id)
    assert model2.dict_attr == {"a": 3, "b": 4}

    model2.dict_attr["c"] = 5
    assert model2.dict_attr == {"a": 3, "b": 4, "c": 5}


def test_uuid_attribute(session: Session) -> None:
    # uuid from string
    model = DummyModel2(uuid="c5ad316a-2cd0-4f78-a49b-cff216c10713")
    session.add(model)
    session.commit()
    model_id = model.id
    # pyre-fixme[16]: `Session` has no attribute `remove`.
    session.remove()
    model2 = DummyModel2.query.get(model_id)

    assert isinstance(model2.uuid, uuid.UUID)

    # plain UUID object
    u = uuid.UUID("3eb7f164-bf15-4564-a058-31bdea0196e6")
    model = DummyModel2(uuid=u)
    session.add(model)
    session.commit()
    model_id = model.id
    # pyre-fixme[16]: `Session` has no attribute `remove`.
    session.remove()
    model2 = DummyModel2.query.get(model_id)

    assert isinstance(model2.uuid, uuid.UUID)
    assert model2.uuid == u



import logging
import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import mkstemp
from typing import Iterator, Optional, Union

logger = logging.getLogger(__name__)

# Hack for Mac OS + homebrew
os.environ["PATH"] += ":/usr/local/bin"

TMP_DIR = "tmp"
CACHE_DIR = "cache"


def get_tmp_dir() -> Path:
    from . import converter

    # pyre-fixme[16]: `Converter` has no attribute `tmp_dir`.
    return converter.tmp_dir


# Utils
@contextmanager
def make_temp_file(
    blob: Optional[bytes] = None,
    prefix: str = "tmp",
    suffix: str = "",
    tmp_dir: Optional[Path] = None,
) -> Iterator[Union[Iterator, Iterator[str]]]:
    if tmp_dir is None:
        tmp_dir = get_tmp_dir()

    fd, filename = mkstemp(dir=str(tmp_dir), prefix=prefix, suffix=suffix)
    if blob is not None:
        fd = os.fdopen(fd, "wb")
        fd.write(blob)
        fd.close()
    else:
        os.close(fd)

    # pyre-fixme[7]: Expected `Iterator[Union[Iterator[Any], Iterator[str]]]` but
    #  got `Generator[str, None, None]`.
    yield filename
    try:
        os.remove(filename)
    except OSError:
        pass



from .base import JSONView, View
from .object import (
    BaseObjectView,
    JSONBaseSearch,
    JSONModelSearch,
    JSONWhooshSearch,
    ObjectCreate,
    ObjectDelete,
    ObjectEdit,
    ObjectView,
)
from .registry import Registry, default_view

__all__ = (
    "View",
    "JSONView",
    "Registry",
    "default_view",
    "BaseObjectView",
    "ObjectView",
    "ObjectEdit",
    "ObjectCreate",
    "ObjectDelete",
    "JSONBaseSearch",
    "JSONModelSearch",
    "JSONWhooshSearch",
)



from abilian.core.extensions import db
from abilian.services import Service

from .models import Hit, View

__all__ = ["viewtracker"]


class ViewTracker(Service):
    name = "viewtracker"

    @staticmethod
    def record_hit(entity, user):
        # Using user_id here in case user is a threadload proxy
        views = View.query.filter(View.entity == entity, View.user_id == user.id)
        if views.count():
            view = views.first()
            hit = Hit(view=view)
            db.session.add(hit)

        else:
            view = View(entity=entity, user_id=user.id)
            hit = Hit(view=view)
            db.session.add(view)
            db.session.add(hit)

        # FIXME: not the right place
        db.session.commit()

    @staticmethod
    def get_views(entity=None, entities=None, user=None, users=None):
        assert entity is None or entities is None
        assert entity is not None or entities is not None
        assert user is None or users is None

        if entities is None:
            entities = []
        if entity:
            entities += [entity]

        if users is None:
            users = []
        if user is not None:
            users += [user]

        query = View.query
        if entities:
            entity_ids = [entity_.id for entity_ in entities]
            query = query.filter(View.entity_id.in_(entity_ids))

        if users:
            user_ids = [user_.id for user_ in users]
            query = query.filter(View.user_id.in_(user_ids))

        return query.all()

    @staticmethod
    def get_hits(views=None, view=None):
        assert view is not None or views is not None

        if views is None:
            views = []
        if view is not None:
            views += [view]

        view_ids = [view_.id for view_ in views]
        return (
            Hit.query.filter(Hit.view_id.in_(view_ids))
            .order_by(Hit.viewed_at.asc())
            .all()
        )


viewtracker = ViewTracker()



# -*- coding: utf-8 -*-
"""View decorators for controlling some aspect of HTTP, mainly: Cache
headers."""
from functools import wraps

from flask import make_response


def nocache(view):
    @wraps(view)
    def _nocache(*args, **kwargs):
        resp = make_response(view(*args, **kwargs))
        resp.headers["Cache-Control"] = "no-cache"
        return resp

    return _nocache



""""""
import html

import sqlalchemy as sa
import sqlalchemy.orm
from flask import render_template_string, request
from sqlalchemy.sql.expression import asc, desc, func, nullslast

from abilian.core.models.subjects import Group, User
from abilian.i18n import _l
from abilian.services import get_service
from abilian.services.security.models import Role
from abilian.web.action import ButtonAction, FAIcon
from abilian.web.nav import BreadcrumbItem
from abilian.web.util import url_for
from abilian.web.views import base
from abilian.web.views import object as views

from .forms import GroupAdminForm


class JsonGroupsList(base.JSONView):
    """JSON group list for datatable."""

    def data(self, *args, **kw):
        security = get_service("security")
        length = int(kw.get("iDisplayLength", 0))
        start = int(kw.get("iDisplayStart", 0))
        sort_dir = kw.get("sSortDir_0", "asc")
        echo = int(kw.get("sEcho", 0))
        search = kw.get("sSearch", "").replace("%", "").strip().lower()

        end = start + length
        query = Group.query.options(sa.orm.noload("*"))
        total_count = query.count()

        if search:
            # TODO: gérer les accents
            query = query.filter(func.lower(Group.name).like("%" + search + "%"))

        count = query.count()
        columns = [func.lower(Group.name)]
        direction = asc if sort_dir == "asc" else desc
        order_by = list(map(direction, columns))

        # sqlite does not support 'NULLS FIRST|LAST' in ORDER BY clauses
        engine = query.session.get_bind(Group.__mapper__)
        if engine.name != "sqlite":
            order_by[0] = nullslast(order_by[0])

        query = query.order_by(*order_by).add_columns(Group.members_count)
        groups = query.slice(start, end).all()
        data = []

        for group, members_count in groups:
            # TODO: this should be done on the browser.
            group_url = url_for(".groups_group", group_id=group.id)
            name = html.escape(group.name or "")
            roles = [r for r in security.get_roles(group) if r.assignable]

            columns = [
                f'<a href="{group_url}">{name}</a>',
                str(members_count or 0),
                render_template_string(
                    """{%- for role in roles %}
                        <span class="badge badge-default">{{ role }}</span>
                        {%- endfor %}""",
                    roles=roles,
                ),
                "\u2713" if group.public else "",
            ]

            data.append(columns)

        return {
            "sEcho": echo,
            "iTotalRecords": total_count,
            "iTotalDisplayRecords": count,
            "aaData": data,
        }


# Group edit / create views
class GroupBase:
    Model = Group
    Form = GroupAdminForm
    pk = "group_id"
    base_template = "admin/_base.html"

    def index_url(self):
        return url_for(".groups")

    def view_url(self):
        return url_for(".groups_group", group_id=self.obj.id)


# those buttons are made to have valid edit actions, but will not be shown in
# edit forms: they must be availabe only during POST
ADD_USER_BUTTON = ButtonAction(
    "form",
    "add_user",
    condition=lambda v: request.method == "POST",
    title=_l("Add"),
    btn_class="primary",
)

REMOVE_USER_BUTTON = ButtonAction(
    "form",
    "remove_user",
    condition=lambda v: request.method == "POST",
    btn_class="danger",
    icon=FAIcon("times"),
    title="",
)


class GroupView(GroupBase, views.ObjectView):
    template = "admin/group_view.html"

    def breadcrumb(self):
        label = render_template_string("<em>{{ g }}</em>", g=self.obj.name)
        return BreadcrumbItem(label=label, url="", description=self.obj.name)

    @property
    def template_kwargs(self):
        security = get_service("security")
        kw = super().template_kwargs
        members = list(self.obj.members)
        members.sort(key=lambda u: (u.last_name, u.first_name))
        kw["members"] = members
        kw["roles"] = sorted(
            r for r in security.get_roles(self.obj, no_group_roles=True) if r.assignable
        )
        kw["ADD_USER_BUTTON"] = ADD_USER_BUTTON
        kw["REMOVE_USER_BUTTON"] = REMOVE_USER_BUTTON
        return kw


class GroupEdit(GroupBase, views.ObjectEdit):
    def breadcrumb(self):
        label = render_template_string("<em>{{ g }}</em>", g=self.obj.name)
        return BreadcrumbItem(label=label, url="", description=self.obj.name)

    def get_form_buttons(self, *args, **kwargs):
        buttons = super().get_form_buttons()
        buttons.append(ADD_USER_BUTTON)
        buttons.append(REMOVE_USER_BUTTON)
        return buttons

    def get_form_kwargs(self):
        kw = super().get_form_kwargs()
        security = get_service("security")
        roles = [
            r for r in security.get_roles(self.obj, no_group_roles=True) if r.assignable
        ]
        kw["roles"] = [role.name for role in roles]
        return kw

    def after_populate_obj(self):
        security = get_service("security")
        current_roles = security.get_roles(self.obj, no_group_roles=True)
        current_roles = {r for r in current_roles if r.assignable}
        new_roles = {Role(r) for r in self.form.roles.data}

        for r in current_roles - new_roles:
            security.ungrant_role(self.obj, r)

        for r in new_roles - current_roles:
            security.grant_role(self.obj, r)

        return super().after_populate_obj()

    def add_user(self, *args, **kwargs):
        user_id = int(request.form.get("user"))
        user = User.query.options(sa.orm.joinedload(User.groups)).get(user_id)
        self.obj.members.add(user)
        sa.orm.object_session(self.obj).commit()
        return self.redirect_to_view()

    def remove_user(self, *args, **kwargs):
        user_id = int(request.form.get("user"))
        user = User.query.options(sa.orm.joinedload(User.groups)).get(user_id)
        self.obj.members.discard(user)
        sa.orm.object_session(self.obj).commit()
        return self.redirect_to_view()


class GroupCreate(GroupBase, views.ObjectCreate):
    chain_create_allowed = True



"""Class based views."""
import logging

import sqlalchemy as sa
import sqlalchemy.exc
import sqlalchemy.orm
from flask import current_app, flash, g, redirect, render_template, request, \
    url_for
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.entities import ValidationError
from abilian.core.extensions import db
from abilian.core.signals import activity
from abilian.i18n import _, _l
from abilian.services import get_service
from abilian.services.security import CREATE, DELETE, READ, WRITE

from .. import csrf, forms, nav
from ..action import ButtonAction, Endpoint, actions
from .base import JSONView, View

logger = logging.getLogger(__name__)


class BaseObjectView(View):
    """Base class common to all database objects views."""

    #: form title
    title = None

    #: Model class
    Model = None

    #: primary key name to look for in url arguments
    pk = "object_id"

    #: object instance for this view
    obj = None

    #: object id
    object_id = None

    #: template to render
    template = None

    #: default templates inherit from "base_template". This allows to use
    #: generic templates with a custom base
    base_template = "base.html"

    def __init__(self, Model=None, pk=None, base_template=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls = self.__class__
        self.pk = pk if pk is not None else cls.pk
        self.Model = Model if Model is not None else cls.Model
        self.base_template = (
            base_template if base_template is not None else cls.base_template
        )

    def prepare_args(self, args, kwargs):
        args, kwargs = self.init_object(args, kwargs)

        if self.obj is None:
            raise NotFound()

        return args, kwargs

    def breadcrumb(self):
        """Return :class:`..nav.BreadcrumbItem` instance for this object.

        This method may return a list of BreadcrumbItem instances.
        Return `None` if nothing.
        """
        return None

    def init_object(self, args, kwargs):
        """This method is reponsible for setting :attr:`obj`.

        It is called during :meth:`prepare_args`.
        """
        self.object_id = kwargs.pop(self.pk, None)
        if self.object_id is not None:
            self.obj = self.Model.query.get(self.object_id)
            actions.context["object"] = self.obj

        return args, kwargs

    def get(self, *args, **kwargs):
        bc = self.breadcrumb()
        if bc is not None:
            bc = [bc] if isinstance(bc, nav.BreadcrumbItem) else list(bc)
            assert all(isinstance(b, nav.BreadcrumbItem) for b in bc)
            g.breadcrumb.extend(bc)

        kwargs = {"base_template": self.base_template}
        kwargs.update(self.template_kwargs)
        # forbid override "view"
        kwargs["view"] = self
        return render_template(self.template, **kwargs)

    @property
    def template_kwargs(self):
        """Get template render arguments.

        You may override `base_template` for instance. Only `view`
        cannot be overriden.
        """
        return {}


class ObjectView(BaseObjectView):
    """View objects."""

    #: html template
    template = "default/object_view.html"

    #: View form class. Form object used to show objects fields
    Form = None

    #: required permission. Must be an instance of
    #: :class:`abilian.services.security.Permission`
    permission = READ

    #: form instance for this view
    form = None

    def __init__(self, Model=None, pk=None, Form=None, template=None, *args, **kwargs):
        super().__init__(Model, pk, *args, **kwargs)
        cls = self.__class__
        self.Form = Form if Form is not None else cls.Form
        self.template = template if template is not None else cls.template

    def prepare_args(self, args, kwargs):
        """
        :attr:`form` is initialized here. See also :meth:`View.prepare_args`.
        """
        args, kwargs = super().prepare_args(args, kwargs)
        form_kwargs = self.get_form_kwargs()
        self.form = self.Form(**form_kwargs)
        return args, kwargs

    def get_form_kwargs(self):
        kw = {"obj": self.obj}
        if issubclass(self.Form, forms.Form) and self.permission:
            kw["permission"] = self.permission
        return kw

    def index_url(self):
        return url_for(".index")

    def redirect_to_index(self):
        return redirect(self.index_url())

    @property
    def template_kwargs(self):
        """Provides :attr:`form` to templates."""
        kw = super().template_kwargs
        kw["form"] = self.form
        return kw


CANCEL_BUTTON = ButtonAction(
    "form",
    "cancel",
    title=_l("Cancel"),
    # .cancel: if jquery.validate is used it will properly skip validation
    btn_class="default cancel",
)

EDIT_BUTTON = ButtonAction("form", "edit", btn_class="primary", title=_l("Save"))

ADD_ANOTHER_BUTTON = ButtonAction(
    "form",
    "create_add_another",
    btn_class="primary",
    title=_l("Create and add another"),
    condition=lambda ctx: getattr(ctx["view"], "add_another_button", False),
)


class ObjectEdit(ObjectView):
    """Edit object."""

    template = "default/object_edit.html"
    decorators = (csrf.support_graceful_failure,)
    permission = WRITE

    #: :class:ButtonAction instance to show on form
    _buttons = ()

    #: submitted form data
    data = None

    #: action name from form data
    action = None

    #: button clicked, corresponding to :attr:`action`.
    button = None

    #: verb used to describe activity
    activity_verb = "update"

    #: UI flash message
    _message_success = _l("Entity successfully edited")

    view_endpoint = None

    def __init__(
        self,
        Model=None,
        pk=None,
        Form=None,
        template=None,
        view_endpoint=None,
        message_success=None,
        *args,
        **kwargs,
    ):
        super().__init__(Model, pk, Form, template=template, *args, **kwargs)
        if view_endpoint is not None:
            self.view_endpoint = view_endpoint

        if not self.view_endpoint:
            self.view_endpoint = f".{self.Model.__name__}_view"

        if message_success:
            self._message_success = message_success

    def post(self, *args, **kwargs):
        # conservative: no action submitted -> cancel
        action = self.data.get("__action", "cancel")
        if action == "cancel":
            return self.cancel()

        return self.handle_action(action)

    def put(self):
        return self.post()

    def prepare_args(self, args, kwargs):
        args, kwargs = super().prepare_args(args, kwargs)
        self._buttons = self.get_form_buttons(*args, **kwargs)
        self.data = request.form
        return args, kwargs

    def get_form_buttons(self, *args, **kwargs):
        return [EDIT_BUTTON, CANCEL_BUTTON]

    @property
    def buttons(self):
        return (button for button in self._buttons if button.available(actions.context))

    def view_url(self):
        kw = {self.pk: self.obj.id}
        return url_for(self.view_endpoint, **kw)

    def redirect_to_view(self):
        if self.button:
            url = self.button.url(actions.context)
            if url:
                return redirect(url)
        return redirect(self.view_url())

    def message_success(self):
        return str(self._message_success)

    # actions
    def handle_action(self, action):
        for button in self._buttons:
            if action == button.name:
                if not button.available({"view": self}):
                    raise ValueError(
                        'Action "{}" not available' "".format(action.encode("utf-8"))
                    )
                break
        else:
            raise ValueError('Unknown action: "{}"'.format(action.encode("utf-8")))

        self.action = action
        self.button = button
        return getattr(self, action)()

    def cancel(self):
        return self.redirect_to_view()

    def edit(self, redirect_to=None):
        if self.validate():
            return self.form_valid(redirect_to=redirect_to)
        else:
            if request.csrf_failed:
                errors = self.form.errors
                csrf_failed = errors.pop("csrf_token", False)
                if csrf_failed and not errors:
                    # failed only because of invalid/expired csrf, no error on
                    # form
                    return self.form_csrf_invalid()

            resp = self.form_invalid()
            if resp:
                return resp

            flash(_("Please fix the error(s) below"), "error")

        # if we end here then something wrong has happened: show form with error
        # messages
        return self.get()

    def before_populate_obj(self):
        """This method is called after form has been validated and before
        calling `form.populate_obj()`.

        Sometimes one may want to remove a field
        from the form because it's non-sense to store it on edited object, and
        use it in a specific manner, for example::

            image = form.image
            del form.image
            store_image(image)
        """
        pass

    def after_populate_obj(self):
        """Called after `self.obj` values have been updated, and `self.obj`
        attached to an ORM session."""
        pass

    def handle_commit_exception(self, exc):
        """Hook point to handle exception that may happen during commit.

        It is the responsability of this method to perform a rollback if it is
        required for handling `exc`. If the method does not handle `exc` if should
        do nothing and return None.

        :returns: * a valid :class:`Response` if exception is handled.
                  * `None` if exception is not handled. Default handling happens.
        """
        return None

    def commit_success(self):
        """Called after object has been successfully saved to database."""

    def validate(self):
        return self.form.validate()

    def form_valid(self, redirect_to=None):
        """Save object.

        Called when form is validated.

        :param redirect_to: real url (created with url_for) to redirect to,
          instead of the view by default.
        """
        session = db.session()

        with session.no_autoflush:
            self.before_populate_obj()
            self.form.populate_obj(self.obj)
            session.add(self.obj)
            self.after_populate_obj()

        try:
            session.flush()
            self.send_activity()
            session.commit()

        except ValidationError as e:
            rv = self.handle_commit_exception(e)
            if rv is not None:
                return rv
            session.rollback()
            flash(str(e), "error")
            return self.get()

        except sa.exc.IntegrityError as e:
            rv = self.handle_commit_exception(e)
            if rv is not None:
                return rv
            session.rollback()
            logger.error(e)
            flash(_("An entity with this name already exists in the system."), "error")
            return self.get()

        else:
            self.commit_success()
            flash(self.message_success(), "success")

            if redirect_to:
                return redirect(redirect_to)
            else:
                return self.redirect_to_view()

    def form_invalid(self):
        """When a form doesn't validate this method is called.

        It may return a :class:`Flask.Response` instance, to handle specific
        errors in custom screens.

        Else the edit form screen is returned with error(s) highlighted.

        This method is useful for detecting edition conflict using hidden fields
        and show a specific screen to help resolve the conflict.
        """
        return None

    def form_csrf_invalid(self):
        """Called when a form doesn't validate *only* because of csrf token
        expiration.

        This works only if form is an instance of :class:`flask_wtf.form.SecureForm`.
        Else default CSRF protection (before request) will take place.

        It must return a valid :class:`Flask.Response` instance. By default it
        returns to edit form screen with an informative message.
        """
        current_app.extensions["csrf-handler"].flash_csrf_failed_message()
        return self.get()

    def send_activity(self):
        activity.send(
            self,
            actor=g.user,
            verb=self.activity_verb,
            object=self.obj,
            target=self.activity_target,
        )

    @property
    def activity_target(self):
        """Return `target` to use when creating activity."""
        return None


CREATE_BUTTON = ButtonAction("form", "create", btn_class="primary", title=_l("Create"))
CHAIN_CREATE_BUTTON = ButtonAction(
    "form",
    "chain_create",
    btn_class="primary",
    title=_l("Create and add new"),
    # pyre-fixme[6]: Expected `str` for 1st param but got `Optional[str]`.
    endpoint=lambda ctx: Endpoint(request.endpoint, **request.view_args),
    condition=lambda ctx: getattr(ctx["view"], "chain_create_allowed", False),
)


class ObjectCreate(ObjectEdit):
    """Create a new object."""

    permission = CREATE
    activity_verb = "post"
    _message_success = _l("Entity successfully added")

    #: set to `True` to show 'Save and add new' button
    chain_create_allowed = False

    def __init__(self, chain_create_allowed=None, *args, **kwargs):
        if chain_create_allowed is not None:
            self.chain_create_allowed = bool(chain_create_allowed)

        super().__init__(*args, **kwargs)

    def prepare_args(self, args, kwargs):
        # we must ensure that no flush() occurs and that obj is not registered
        # in session (to prevent accidental insert of an incomplete object)
        session = db.session()
        with session.no_autoflush:
            args, kwargs = super().prepare_args(args, kwargs)

        try:
            session.expunge(self.obj)
        except sa.exc.InvalidRequestError:
            # obj is not in session
            pass

        return args, kwargs

    def init_object(self, args, kwargs):
        self.obj = self.Model()
        return args, kwargs

    def get_form_kwargs(self):
        kw = super().get_form_kwargs()
        if request.method == "GET":
            # when GET allow form prefill instead of empty/current object data
            # FIXME: filter allowed parameters on given a field flags (could be
            # 'allow_from_get'?)
            kw["formdata"] = request.args

        return kw

    def get_form_buttons(self, *args, **kwargs):
        return [CREATE_BUTTON, CHAIN_CREATE_BUTTON, CANCEL_BUTTON]

    def breadcrumb(self):
        return nav.BreadcrumbItem(label=CREATE_BUTTON.title)

    # actions
    def create(self):
        return self.edit()

    chain_create = create

    def cancel(self):
        return self.redirect_to_index()


DELETE_BUTTON = ButtonAction("form", "delete", title=_l("Delete"))


class ObjectDelete(ObjectEdit):
    """Delete object.

    Supports the DELETE verb.
    """

    methods = ["POST"]
    permission = DELETE
    activity_verb = "delete"
    _message_success = _l("Entity deleted")

    init_object = BaseObjectView.init_object

    def get_form_buttons(self, *args, **kwargs):
        return [DELETE_BUTTON, CANCEL_BUTTON]

    def delete(self):
        session = db.session()
        session.delete(self.obj)
        activity.send(
            self,
            actor=g.user,
            verb="delete",
            object=self.obj,
            target=self.activity_target,
        )
        try:
            session.commit()
        except sa.exc.IntegrityError as e:
            rv = self.handle_commit_exception(e)
            if rv is not None:
                return rv
            session.rollback()
            logger.error(e)
            flash(
                _("This entity is referenced by another object and cannot be deleted."),
                "error",
            )
            return self.redirect_to_view()
        else:
            flash(self.message_success(), "success")
            # FIXME: for DELETE verb response in case of success should be 200, 202
            # (accepted) or 204 (no content)
            return self.redirect_to_index()


class JSONBaseSearch(JSONView):
    Model = None
    minimum_input_length = 2

    def __init__(self, *args, **kwargs):
        Model = kwargs.pop("Model", self.Model)
        minimum_input_length = kwargs.pop(
            "minimum_input_length", self.minimum_input_length
        )
        super().__init__(*args, **kwargs)
        self.Model = Model
        self.minimum_input_length = minimum_input_length

    def prepare_args(self, args, kwargs):
        args, kwargs = JSONView.prepare_args(self, args, kwargs)
        kwargs["q"] = kwargs.get("q", "").replace("%", " ").lower()
        return args, kwargs

    def data(self, q, *args, **kwargs):
        if self.minimum_input_length and len(q) < self.minimum_input_length:
            msg = f"Minimum query length is {self.minimum_input_length:d}"
            raise BadRequest(msg)

        results = []
        for obj in self.get_results(q, **kwargs):
            results.append(self.get_item(obj))

        return {"results": results}

    def get_results(self, q, *args, **kwargs):
        raise NotImplementedError

    def get_item(self, obj):
        """Return a result item.

        :param obj: Instance object
        :returns: a dictionnary with at least `id` and `text` values
        """
        raise NotImplementedError


class JSONModelSearch(JSONBaseSearch):
    """Base class for json sqlalchemy model search.

    As used by select2 widgets for example.
    """

    def get_results(self, q, *args, **kwargs):
        query = self.Model.query

        query = self.options(query)
        query = self.filter(query, q, **kwargs)
        query = self.order_by(query)
        if not q and not self.minimum_input_length:
            query = query.limit(50)
        return query.all()

    def options(self, query):
        return query.options(sa.orm.noload("*"))

    def filter(self, query, q, **kwargs):
        if not q:
            return query
        return query.filter(sa.func.lower(self.Model.name).like(q + "%"))

    def order_by(self, query):
        return query.order_by(self.Model.name)

    def get_label(self, obj):
        return obj.name

    def get_item(self, obj):
        """Return a result item.

        :param obj: Instance object
        :returns: a dictionnary with at least `id` and `text` values
        """
        return {"id": obj.id, "text": self.get_label(obj), "name": obj.name}


class JSONWhooshSearch(JSONBaseSearch):
    """Base class for JSON Whoosh search, as used by select2 widgets for
    example."""

    def get_results(self, q, *args, **kwargs):
        svc = get_service("indexing")
        search_kwargs = {"limit": 50, "Models": (self.Model,)}
        results = svc.search(q, **search_kwargs)

        itemkey = None
        try:
            # 'nom' doesn't always exist but for Contacts, sorting on
            # the last name ('nom') feels more natural than 'name',
            # which starts with the first name ('prenom').
            if not results.is_empty():
                res = results[0]
                fields = res.fields()
                if "nom" in fields:
                    itemkey = "nom"
                elif "name" in fields:
                    itemkey = "name"
                if itemkey:
                    results.sort(key=lambda it: it.fields().get(itemkey))
        except Exception:
            if itemkey is not None:
                msg = "we could not sort whoosh results on fields' key {}.".format(
                    itemkey
                )
                logger.warning(msg)

        return results

    def get_item(self, hit):
        """Return a result item.

        :param hit: Hit object from Whoosh
        :returns: a dictionnary with at least `id` and `text` values
        """
        return {"id": hit["id"], "text": hit["name"], "name": hit["name"]}



from .models import (
    CREATE,
    DELETE,
    MANAGE,
    READ,
    WRITE,
    Admin,
    Anonymous,
    Authenticated,
    Creator,
    InheritSecurity,
    Manager,
    Owner,
    Permission,
    PermissionAssignment,
    Reader,
    Role,
    RoleAssignment,
    RoleType,
    SecurityAudit,
    Writer,
)
from .service import security, SecurityService



""""""
import json
import logging
import time
from datetime import timedelta
from typing import Any, Dict
from uuid import UUID, uuid1

from celery import shared_task
from flask import current_app
from flask_login import AnonymousUserMixin

from abilian.app import Application
from abilian.core import signals
from abilian.web import url_for

from .views import bp as blueprint

logger = logging.getLogger(__name__)

CHUNK_SIZE = 64 * 1024

DEFAULT_CONFIG = {
    "USER_QUOTA": 100 * 1024 ** 2,  # max 100 Mb for all current files
    "USER_MAX_FILES": 1000,  # max number of files per user
    "DELETE_STALLED_AFTER": 60 * 60 * 24,  # delete files remaining after 1 day
}

CLEANUP_SCHEDULE_ID = __name__ + ".periodic_clean_upload_directory"
DEFAULT_CLEANUP_SCHEDULE = {"task": CLEANUP_SCHEDULE_ID, "schedule": timedelta(hours=1)}


def is_valid_handle(handle: str) -> bool:
    try:
        UUID(handle)
    except ValueError:
        return False

    return True


class FileUploadsExtension:
    """API for Out-Of-Band file uploads.

    Allow to manage files in forms: file is uploaded to an upload url, a handle is
    returned will be used in the form to refer to this uploaded filed.

    If the form fails to validate the uploaded file is not lost.

    A periodic task cleans the temporary repository.
    """

    def __init__(self, app: Application) -> None:
        app.extensions["uploads"] = self
        app.add_template_global(self, "uploads")
        app.register_blueprint(blueprint)
        signals.register_js_api.connect(self._do_register_js_api)

        self.config: Dict[str, Any] = {}
        self.config.update(DEFAULT_CONFIG)
        self.config.update(app.config.get("FILE_UPLOADS", {}))
        app.config["FILE_UPLOADS"] = self.config

        # celery schedule
        CELERYBEAT_SCHEDULE = app.config.setdefault("CELERYBEAT_SCHEDULE", {})
        if CLEANUP_SCHEDULE_ID not in CELERYBEAT_SCHEDULE:
            CELERYBEAT_SCHEDULE[CLEANUP_SCHEDULE_ID] = DEFAULT_CLEANUP_SCHEDULE

        path = self.UPLOAD_DIR = app.data_dir / "uploads"
        if not path.exists():
            path.mkdir(mode=0o775, parents=True)

        path.resolve()

    def _do_register_js_api(self, sender: Application) -> None:
        app = sender
        # pyre-fixme[6]: Expected `str` for 2nd param but got `Dict[_KT, _VT]`.
        js_api = app.js_api.setdefault("upload", {})
        # pyre-fixme[16]: `str` has no attribute `__setitem__`.
        js_api["newFileUrl"] = url_for("uploads.new_file")

    def user_dir(self, user):
        if isinstance(user, AnonymousUserMixin):
            user_id = "anonymous"
        else:
            user_id = str(user.id)
        return self.UPLOAD_DIR / user_id

    def add_file(self, user, file_obj, **metadata):
        """Add a new file.

        :returns: file handle
        """
        user_dir = self.user_dir(user)
        if not user_dir.exists():
            user_dir.mkdir(mode=0o775)

        handle = str(uuid1())
        file_path = user_dir / handle

        with file_path.open("wb") as out:
            for chunk in iter(lambda: file_obj.read(CHUNK_SIZE), b""):
                out.write(chunk)

        if metadata:
            meta_file = user_dir / f"{handle}.metadata"
            with meta_file.open("wb") as out:
                metadata_json = json.dumps(metadata, skipkeys=True).encode("ascii")
                out.write(metadata_json)

        return handle

    def get_file(self, user, handle):
        """Retrieve a file for a user.

        :returns: a :class:`pathlib.Path` instance to this file,
            or None if no file can be found for this handle.
        """
        user_dir = self.user_dir(user)
        if not user_dir.exists():
            return None

        if not is_valid_handle(handle):
            return None

        file_path = user_dir / handle

        if not file_path.exists() and not file_path.is_file():
            return None

        return file_path

    def get_metadata_file(self, user, handle):
        content = self.get_file(user, handle)
        if content is None:
            return None

        metafile = content.parent / f"{handle}.metadata"
        if not metafile.exists():
            return None

        return metafile

    def get_metadata(self, user, handle):
        metafile = self.get_metadata_file(user, handle)
        if metafile is None:
            return {}

        try:
            with metafile.open("rb") as in_:
                meta = json.load(in_)
        except Exception:
            meta = {}

        return meta

    def remove_file(self, user, handle):
        paths = (self.get_file(user, handle), self.get_metadata_file(user, handle))

        for file_path in paths:
            if file_path is not None:
                try:
                    file_path.unlink()
                except Exception:
                    logger.exception("Error during remove file")

    def clear_stalled_files(self):
        """Scan upload directory and delete stalled files.

        Stalled files are files uploaded more than
        `DELETE_STALLED_AFTER` seconds ago.
        """
        # FIXME: put lock in directory?
        CLEAR_AFTER = self.config["DELETE_STALLED_AFTER"]
        minimum_age = time.time() - CLEAR_AFTER

        for user_dir in self.UPLOAD_DIR.iterdir():
            if not user_dir.is_dir():
                logger.error("Found non-directory in upload dir: %r", bytes(user_dir))
                continue

            for content in user_dir.iterdir():
                if not content.is_file():
                    logger.error(
                        "Found non-file in user upload dir: %r", bytes(content)
                    )
                    continue

                if content.stat().st_ctime < minimum_age:
                    content.unlink()


# Task scheduled to run every hour: make it expire after 50min.
@shared_task(expires=3000)
def periodic_clean_upload_directory():
    """This task should be run periodically.

    Default config sets up schedule using
    :data:`DEFAULT_CLEANUP_SCHEDULE`. `CELERYBEAT_SCHEDULE` key is
    :data:`CLEANUP_SCHEDULE_ID`.
    """
    uploads = current_app.extensions["uploads"]
    uploads.clear_stalled_files()



"""Base Flask application class, used by tests or to be extended in real
applications."""
import logging
import logging.config
from functools import partial
from pathlib import Path

import sqlalchemy as sa
import yaml
from flask import Flask, _request_ctx_stack, g, render_template
from pkg_resources import resource_filename
from sqlalchemy.orm.attributes import NEVER_SET

from abilian.core import extensions

logger = logging.getLogger(__name__)
db = extensions.db


class ErrorManagerMixin(Flask):
    def setup_logging(self) -> None:
        # Force flask to create application logger before logging
        # configuration; else, flask will overwrite our settings
        self.logger  # noqa

        log_level = self.config.get("LOG_LEVEL")
        if log_level:
            self.logger.setLevel(log_level)

        logging_file = self.config.get("LOGGING_CONFIG_FILE")
        if logging_file:
            logging_file = (Path(self.instance_path) / logging_file).resolve()
        else:
            logging_file = Path(
                resource_filename("abilian.core", "default_logging.yml")
            )

        if logging_file.suffix == ".ini":
            # old standard 'ini' file config
            logging.config.fileConfig(str(logging_file), disable_existing_loggers=False)
        elif logging_file.suffix == ".yml":
            # yaml config file
            logging_cfg = yaml.safe_load(logging_file.open())
            logging_cfg.setdefault("version", 1)
            logging_cfg.setdefault("disable_existing_loggers", False)
            logging.config.dictConfig(logging_cfg)

    def handle_user_exception(self, e):
        # If session.transaction._parent is None, then exception has occured in
        # after_commit(): doing a rollback() raises an error and would hide
        # actual error.
        session = db.session()
        if session.is_active and session.transaction._parent is not None:
            # Inconditionally forget all DB changes, and ensure clean session
            # during exception handling.
            session.rollback()
        else:
            self._remove_session_save_objects()

        return Flask.handle_user_exception(self, e)

    def handle_exception(self, e):
        session = db.session()
        if not session.is_active:
            # Something happened in error handlers and session is not usable
            # anymore.
            self._remove_session_save_objects()

        return Flask.handle_exception(self, e)

    def _remove_session_save_objects(self):
        """Used during exception handling in case we need to remove() session:

        keep instances and merge them in the new session.
        """
        if self.testing:
            return
        # Before destroying the session, get all instances to be attached to the
        # new session. Without this, we get DetachedInstance errors, like when
        # tryin to get user's attribute in the error page...
        old_session = db.session()
        g_objs = []
        for key in iter(g):
            obj = getattr(g, key)
            if isinstance(obj, db.Model) and sa.orm.object_session(obj) in (
                None,
                old_session,
            ):
                g_objs.append((key, obj, obj in old_session.dirty))

        db.session.remove()
        session = db.session()

        for key, obj, load in g_objs:
            # replace obj instance in bad session by new instance in fresh
            # session
            setattr(g, key, session.merge(obj, load=load))

        # refresh `current_user`
        user = getattr(_request_ctx_stack.top, "user", None)
        if user is not None and isinstance(user, db.Model):
            _request_ctx_stack.top.user = session.merge(user, load=load)

    # pyre-fixme[15]: `log_exception` overrides method defined in `Flask`
    #  inconsistently.
    def log_exception(self, exc_info):
        """Log exception only if Sentry is not used (this avoids getting error
        twice in Sentry)."""
        dsn = self.config.get("SENTRY_DSN")
        if not dsn:
            super().log_exception(exc_info)

    def init_sentry(self):
        """Install Sentry handler if config defines 'SENTRY_DSN'."""
        dsn = self.config.get("SENTRY_DSN")
        if not dsn:
            return

        try:
            import sentry_sdk
        except ImportError:
            logger.error(
                'SENTRY_DSN is defined in config but package "sentry-sdk"'
                " is not installed."
            )
            return

        from sentry_sdk.integrations.flask import FlaskIntegration

        sentry_sdk.init(dsn=dsn, integrations=[FlaskIntegration()])

        # ext = Sentry(self, logging=True, level=logging.ERROR)
        # ext.client.tags["app_name"] = self.name
        # ext.client.tags["process_type"] = "web"
        # server_name = str(self.config.get("SERVER_NAME"))
        # ext.client.tags["configured_server_name"] = server_name

    def install_default_handlers(self) -> None:
        for http_error_code in (403, 404, 500):
            self.install_default_handler(http_error_code)

    def install_default_handler(self, http_error_code: int) -> None:
        """Install a default error handler for `http_error_code`.

        The default error handler renders a template named error404.html
        for http_error_code 404.
        """
        logger.debug(
            "Set Default HTTP error handler for status code %d", http_error_code
        )
        handler = partial(self.handle_http_error, http_error_code)
        self.errorhandler(http_error_code)(handler)

    def handle_http_error(self, code, error):
        """Helper that renders `error{code}.html`.

        Convenient way to use it::

           from functools import partial
           handler = partial(app.handle_http_error, code)
           app.errorhandler(code)(handler)
        """
        # 5xx code: error on server side
        if (code // 100) == 5:
            # ensure rollback if needed, else error page may
            # have an error, too, resulting in raw 500 page :-(
            db.session.rollback()

        template = f"error{code:d}.html"
        return render_template(template, error=error), code



""""""


class Taggable:
    """Mixin trait for taggable objects.

    Currently not used.
    """

    pass


class TagService:
    """The tag service."""

    def tag(self, obj, term, user=None):
        """Apply a tag on a taggable object.

        If user is None, uses the current logged in user.
        """

    def untag(self, obj, term, user=None):
        """Remove the given tag from the given object.

        See tag().
        """

    def get_objects_tagged_with(self, term):
        """Returns a list of objects tagged with a given term."""

    def get_tags_applied_on(self, obj):
        """Returns a list of tags applied on a given document."""



""""""
from typing import Any, Callable, Collection, Optional, Tuple, Union

from flask import Blueprint as BaseBlueprint

from abilian.core.models.subjects import User
from abilian.services.security import Anonymous, Role
from abilian.services.security.service import SecurityService


def allow_anonymous(user: User, roles: Collection[Role], **kwargs: Any) -> bool:
    return True


def allow_access_for_roles(roles: Union[Collection[Role], Role]) -> Callable:
    """Access control helper to check user's roles against a list of valid
    roles."""
    if isinstance(roles, Role):
        roles = (roles,)
    valid_roles = frozenset(roles)

    if Anonymous in valid_roles:
        return allow_anonymous

    # FIXME: parameter not used. Why?
    def check_role(user: User, roles: Collection[Role], **kwargs):
        from abilian.services import get_service

        security: SecurityService = get_service("security")
        return security.has_role(user, valid_roles)

    return check_role


class Blueprint(BaseBlueprint):
    """An enhanced :class:`flask.blueprints.Blueprint` with access control
    helpers."""

    def __init__(
        self,
        name: str,
        import_name: str,
        allowed_roles: Union[None, str, Role, Collection[Role]] = None,
        **kwargs: Any
    ) -> None:
        """
        :param allowed_roles: role or list of roles required to access any view in this
            blueprint.
        """
        BaseBlueprint.__init__(self, name, import_name, **kwargs)

        if allowed_roles is not None:
            if isinstance(allowed_roles, str):
                allowed_roles = Role(allowed_roles)

            if isinstance(allowed_roles, Role):
                allowed_roles = (allowed_roles,)
        else:
            allowed_roles = ()

        if allowed_roles:
            self.record_once(
                lambda s: s.app.add_access_controller(
                    self.name, allow_access_for_roles(allowed_roles)
                )
            )

    def allow_any(self, func):
        pass



""""""
from typing import TYPE_CHECKING, Set

from flask_login import AnonymousUserMixin, LoginManager

if TYPE_CHECKING:
    from abilian.core.models.subjects import Group


class AnonymousUser(AnonymousUserMixin):
    def has_role(self, role):
        from abilian.services import get_service

        security = get_service("security")
        return security.has_role(self, role)

    @property
    def groups(self) -> Set["Group"]:
        return set()


login_manager = LoginManager()
login_manager.anonymous_user = AnonymousUser



""""""
import abc
from typing import Any, Dict

import sqlalchemy as sa
import sqlalchemy.event
from sqlalchemy import Column, ForeignKey, Integer, UnicodeText
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import backref, relationship

from abilian.core.entities import Entity

from .blob import Blob

#: name of backref on target :class:`Entity` object
ATTRIBUTE = "__attachments__"


class SupportAttachment(metaclass=abc.ABCMeta):
    pass


def register(cls):
    """Register an :class:`~.Entity` as a attachmentable class.

    Can be used as a class decorator:

    .. code-block:: python

      @attachment.register
      class MyContent(Entity):
          ....
    """
    if not issubclass(cls, Entity):
        raise ValueError("Class must be a subclass of abilian.core.entities.Entity")

    SupportAttachment.register(cls)
    return cls


def supports_attachments(obj):
    """
    :param obj: a class or instance

    :returns: True is obj supports attachments.
    """
    if isinstance(obj, type):
        return issubclass(obj, SupportAttachment)

    if not isinstance(obj, SupportAttachment):
        return False

    if obj.id is None:
        return False

    return True


def for_entity(obj, check_support_attachments=False):
    """Return attachments on an entity."""
    if check_support_attachments and not supports_attachments(obj):
        return []

    return getattr(obj, ATTRIBUTE)


class Attachment(Entity):
    """An Attachment owned by an :class:`Entity`."""

    __auditable_entity__ = ("entity", "attachment", ("id", "name"))

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        # we cannot use super(Attachment, cls): declared_attr happens during
        # class construction. super(cls, cls) could work; as long as `cls`
        # is not a subclass of `Attachment`: it would enter into
        # an infinite loop.
        #
        # Entity.__mapper_args__ calls the descriptor with 'Entity', not `cls`.
        args = Entity.__dict__["__mapper_args__"].fget(cls)
        args["order_by"] = cls.created_at
        return args

    entity_id = Column(Integer, ForeignKey(Entity.id), nullable=False)

    #: owning entity
    entity = relationship(
        Entity,
        lazy="immediate",
        foreign_keys=[entity_id],
        backref=backref(
            ATTRIBUTE,
            lazy="select",
            order_by="Attachment.created_at",
            cascade="all, delete-orphan",
        ),
    )

    blob_id = Column(Integer, sa.ForeignKey(Blob.id), nullable=False)
    #: file. Stored in a :class:`Blob`
    blob = relationship(Blob, cascade="all, delete", foreign_keys=[blob_id])

    description = Column(UnicodeText(), nullable=False, default="", server_default="")

    def __repr__(self):
        class_ = self.__class__
        mod_ = class_.__module__
        classname = class_.__name__
        return "<{}.{} instance at 0x{:x} entity id={!r}>".format(
            mod_, classname, id(self), self.entity_id
        )


@sa.event.listens_for(Attachment, "before_insert", propagate=True)
@sa.event.listens_for(Attachment, "before_update", propagate=True)
def set_attachment_name(mapper, connection, target):
    if target.name:
        return

    blob = target.blob
    if not blob:
        return

    filename = blob.meta.get("filename")
    if filename is not None:
        target.name = filename



"""Draft API for a like service.

Every entity (and even comments) should be likeable.
"""


class Like:
    pass


class LikeService:
    def like(self, object, user=None):
        """Given or current user likes the given likeable object."""

    def dislike(self, object, user=None):
        """Given or current user dislikes the given likeable object."""

    def get_likes_count_on(self, object):
        """Returns the number of likes on the given likeable object."""

    def get_dislikes_count_on(self, object):
        """Returns the number of dislikes on the given likeable object."""

    def get_likes_on(self, object):
        """Returns all the likes (and dislikes) on a given object."""



"""I18n.

To mark strings for translation::

    from abilian.i18n import _
    _(u'message to translate')

Use :data:`_` for gettext, :data:`_l` for lazy_gettext, :data:`_n` for
ngettext.

:data:`Babel extension<babel>` support multiple translation paths. This
allows to add more catalogs to search for translations, in LIFO
order. This feature can be used to override some translations in a
custom application, be providing a catalog with messages to override::

    current_app.extensions['babel'].add_translations('abilian.core')

See :meth:`add_translations<Babel.add_translations>`.

To extract messages to build the message catalog template (.pot), use the
following "`-k`" parameters:

.. code-block:: bash

    $ pybabel extract -F babel.cfg -k "_n:1,2" -k "_l" -o "msg.pot" "src"


This can be made easier by placing in `setup.cfg`:

.. code-block:: ini

    [extract_messages]
    mapping_file = babel.cfg
    keywords = _n:1,2 _l
    output-file = msg.pot
    input-dirs = src


And just type:

.. code-block:: bash

    $ python setup.py extract_messages
"""
import importlib
import os
import re
import unicodedata
from datetime import datetime, tzinfo
from gettext import GNUTranslations
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import flask_babel
import pytz
from babel import Locale
from babel.dates import LOCALTZ, get_timezone, get_timezone_gmt
from babel.localedata import locale_identifiers
from babel.support import Translations as BaseTranslations
from flask import Flask, _request_ctx_stack, current_app, g, render_template, \
    request
from flask_babel import Babel as BabelBase
from flask_babel import LazyString, force_locale, gettext, lazy_gettext, \
    ngettext

__all__ = [
    "_",
    "_l",
    "_n",
    "babel",
    "get_default_locale",
    "gettext",
    "lazy_country_name",
    "lazy_gettext",
    "localeselector",
    "ngettext",
    "render_template_i18n",
    "timezoneselector",
    "VALID_LANGUAGES_CODE",
]

#: gettext alias
_ = gettext

#: lazy_gettext alias
_l = lazy_gettext
#: ngettext alias
_n = ngettext

#: accepted languages codes
VALID_LANGUAGES_CODE = frozenset(
    lang for lang in locale_identifiers() if len(lang) == 2
)


def get_default_locale():
    return current_app.extensions["babel"].default_locale


def _get_locale() -> Locale:
    locale = flask_babel.get_locale()
    if locale is None:
        locale = get_default_locale()
    return locale


def __gettext_territory(code):
    locale = _get_locale()
    return locale.territories.get(code) or get_default_locale().territories.get(code)


#: get localized territory name
def country_name(code):
    return __gettext_territory(code)


#: lazy version of :func:`country_name`
def lazy_country_name(code):
    return LazyString(__gettext_territory, code)


def default_country() -> str:
    return current_app.config.get("DEFAULT_COUNTRY", "")


def country_choices(
    first: str = "", default_country_first: bool = True
) -> List[Tuple[str, str]]:
    """Return a list of (code, countries), alphabetically sorted on localized
    country name.

    :param first: Country code to be placed at the top
    :param default_country_first:
    :type default_country_first: bool
    """
    locale = _get_locale()
    territories = [
        (code, name) for code, name in locale.territories.items() if len(code) == 2
    ]  # skip 3-digit regions

    if first == "" and default_country_first:
        first = default_country()

    def sortkey(item: Tuple[str, str]) -> str:
        if first is not None and item[0] == first:
            return "0"
        return to_lower_ascii(item[1])

    territories.sort(key=sortkey)
    return territories


def supported_app_locales():
    """Language codes and labels supported by current application.

    :return: an iterable of `(:class:`babel.Locale`, label)`, label being the
    locale language human name in current locale.
    """
    locale = _get_locale()
    codes = current_app.config["BABEL_ACCEPT_LANGUAGES"]
    return ((Locale.parse(code), locale.languages.get(code, code)) for code in codes)


def timezones_choices():
    """Timezones values and their labels for current locale.

    :return: an iterable of `(code, label)`, code being a timezone code and label
    the timezone name in current locale.
    """
    utcnow = pytz.utc.localize(datetime.utcnow())
    locale = _get_locale()
    for tz in sorted(pytz.common_timezones):
        tz = get_timezone(tz)
        now = tz.normalize(utcnow.astimezone(tz))
        label = "({}) {}".format(get_timezone_gmt(now, locale=locale), tz.zone)
        yield (tz, label)  # get_timezone_name(tz, locale=locale))


class Babel(BabelBase):
    """Allow to load translations from other modules."""

    _translations_paths: List[Tuple[str, str]]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init_app(self, app: Flask) -> None:
        super().init_app(app)
        assert app.root_path
        self._translations_paths = [
            (os.path.join(app.root_path, "translations"), "messages")
        ]

    def add_translations(
        self,
        module_name: str,
        translations_dir: str = "translations",
        domain: str = "messages",
    ) -> None:
        """Add translations from external module.

        For example::

            babel.add_translations('abilian.core')

        Will add translations files from `abilian.core` module.
        """
        module = importlib.import_module(module_name)
        # pyre-fixme[16]: `ModuleType` has no attribute `__path__`.
        for path in (Path(p, translations_dir) for p in module.__path__):
            if not (path.exists() and path.is_dir()):
                continue

            if not os.access(str(path), os.R_OK):
                self.app.logger.warning(
                    "Babel translations: read access not allowed {}, skipping."
                    "".format(repr(str(path).encode("utf-8")))
                )
                continue

            self._translations_paths.append((str(path), domain))


class Translations(BaseTranslations):
    """Merge only non-empty translations.

    This avoids having uncomplete catalog that "clear" existing
    translations, when used with :func:`_get_translations_multi_paths`.
    """

    def merge(self, translations: "Translations") -> "Translations":
        if isinstance(translations, GNUTranslations):

            for msgkey, msgstr in translations._catalog.items():
                msgid = msgkey

                if isinstance(msgkey, tuple):
                    msgid = msgkey[0]

                msgstr = msgstr.strip()
                if msgkey in self._catalog and (msgid == msgstr):
                    # when msgstr is empty, compile_catalog sets msgstr = msgid
                    # so this is probable an existing translation that would
                    # be "erased" by msgid string: skip it.

                    # logger.debug('Catalog: %r, skip msgkey: %r, existing: %r',
                    # translations, msgkey, self._catalog[msgkey])
                    continue

                self._catalog[msgkey] = msgstr

            if isinstance(translations, BaseTranslations):
                self.files.extend(translations.files)

        return self


def _get_translations_multi_paths() -> Optional[Translations]:
    """Return the correct gettext translations that should be used for this
    request.

    This will never fail and return a dummy translation object if used
    outside of the request or if a translation cannot be found.
    """
    # pyre-fixme[16]: Module `flask` has no attribute `_request_ctx_stack`.
    ctx = _request_ctx_stack.top
    if ctx is None:
        return None

    translations = getattr(ctx, "babel_translations", None)
    if translations is None:
        babel_ext = ctx.app.extensions["babel"]
        translations = None
        trs = None

        # reverse order: thus the application catalog is loaded last, so that
        # translations from libraries can be overriden
        for (dirname, domain) in reversed(babel_ext._translations_paths):
            trs = Translations.load(
                dirname, locales=[flask_babel.get_locale()], domain=domain
            )

            # babel.support.Translations is a subclass of
            # babel.support.NullTranslations, so we test if object has a 'merge'
            # method

            if not trs or not hasattr(trs, "merge"):
                # got None or NullTranslations instance
                continue
            elif translations is not None and hasattr(translations, "merge"):
                translations.merge(trs)
            else:
                translations = trs

        # ensure translations is at least a NullTranslations object
        if translations is None:
            translations = trs

        ctx.babel_translations = translations

    return translations


# monkey patch flask-babel
flask_babel.get_translations = _get_translations_multi_paths

#: importable instance of :class:`Babel`
babel = Babel()


def localeselector() -> Optional[str]:
    """Default locale selector used in abilian applications."""
    # if a user is logged in, use the locale from the user settings
    user = getattr(g, "user", None)
    if user is not None:
        locale = getattr(user, "locale", None)
        if locale:
            return locale

    # Otherwise, try to guess the language from the user accept header the browser
    # transmits.  By default we support en/fr. The best match wins.
    return request.accept_languages.best_match(
        current_app.config["BABEL_ACCEPT_LANGUAGES"]
    )


def timezoneselector() -> tzinfo:
    """Default timezone selector used in abilian applications."""
    return LOCALTZ


def get_template_i18n(template_name: str, locale: Locale) -> List[str]:
    """Build template list with preceding locale if found."""
    if locale is None:
        return [template_name]

    template_list = []
    parts = template_name.rsplit(".", 1)
    root = parts[0]
    suffix = parts[1]

    if locale.territory is not None:
        locale_string = "_".join([locale.language, locale.territory])
        localized_template_path = ".".join([root, locale_string, suffix])
        template_list.append(localized_template_path)

    localized_template_path = ".".join([root, locale.language, suffix])
    template_list.append(localized_template_path)

    # append the default
    template_list.append(template_name)
    return template_list


class ensure_request_context:
    """Context manager that ensures a request context is set up."""

    _rq_ctx = None

    def __enter__(self) -> None:
        # pyre-fixme[16]: Module `flask` has no attribute `_request_ctx_stack`.
        if _request_ctx_stack.top is None:
            ctx = self._rq_ctx = current_app.test_request_context()
            ctx.__enter__()

    def __exit__(self, *args: Any) -> None:
        ctx = self._rq_ctx
        self._rq_ctx = None

        if ctx is not None:
            ctx.__exit__(*args)


def render_template_i18n(template_name_or_list: str, **context: Any) -> str:
    """Try to build an ordered list of template to satisfy the current
    locale."""
    template_list: List[str] = []
    # Use locale if present in **context
    if "locale" in context:
        locale = Locale.parse(context["locale"])
    else:
        # Use get_locale() or default_locale
        locale = flask_babel.get_locale()

    if isinstance(template_name_or_list, str):
        template_list = get_template_i18n(template_name_or_list, locale)
    else:
        # Search for locale for each member of the list, do not bypass
        for template in template_name_or_list:
            template_list.extend(get_template_i18n(template, locale))

    with ensure_request_context(), force_locale(locale):
        return render_template(template_list, **context)


_NOT_WORD_RE = re.compile(r"[^\w\s]+", flags=re.UNICODE)


def to_lower_ascii(value: str) -> str:
    value = str(value)
    value = _NOT_WORD_RE.sub(" ", value)
    value = unicodedata.normalize("NFKD", value)
    value_b = value.encode("ascii", "ignore")
    value = value_b.decode("ascii")
    value = value.strip().lower()
    value = re.sub(r"[_\s]+", " ", value)
    return value
